/*
 * alias.c Handles command aliases for irc.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990, 1995 Michael Sandroff and others 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: alias.c,v 1.25 1994/07/26 12:12:10 mrg Exp $";
#endif

#include "irc.h"

#include "alias.h"
#include "array.h"
#include "dcc.h"
#include "edit.h"
#include "files.h"
#include "history.h"
#include "hook.h"
#include "input.h"
#include "ircaux.h"
#include "names.h"
#include "output.h"
#include "parse.h"
#include "screen.h"
#include "server.h"
#include "status.h"
#include "vars.h"
#include "window.h"

/* move this to alias.h! */
extern	int	parse_number _((char **));
/*static	char	*evaluate_operator _((char *, char *, char *, char *, int *));*/
static	char	*next_unit _((char *, char *, int *, int));
static	unsigned long	randm _((unsigned long));
static	void 	flush_variables _((void));
static	void 	flush_aliases2 _((void));

static	char	*alias_detected _((void));
static	char	*alias_sent_nick _((void));
static	char	*alias_recv_nick _((void));
static	char	*alias_msg_body _((void));
static	char	*alias_joined_nick _((void));
static	char	*alias_public_nick _((void));
static	char	*alias_dollar _((void));
static	char	*alias_channel _((void));
static	char	*alias_server _((void));
static	char	*alias_query_nick _((void));
static	char	*alias_target _((void));
static	char	*alias_nick _((void));
static	char	*alias_invite _((void));
static	char	*alias_cmdchar _((void));
static	char	*alias_line _((void));
static	char	*alias_away _((void));
static	char	*alias_oper _((void));
static	char	*alias_chanop _((void));
static	char	*alias_modes _((void));
static	char	*alias_buffer _((void));
static	char	*alias_time _((void));
static	char	*alias_version _((void));
static	char	*alias_currdir _((void));
static	char	*alias_current_numeric _((void));
static	char	*alias_server_version _((void));
static  char	*alias_show_userhost _((void));
static	char	*alias_show_realname _((void));
static	char	*alias_online _((void));
static	char 	*alias_idle _((void));
static	char	*alias_version_str _((void));

typedef struct
{
	char	name;
	char	*(*func) _((void));
}	BuiltIns;

static	BuiltIns built_in[] =
{
	{ '.',		alias_sent_nick },
	{ ',',		alias_recv_nick },
	{ ':',		alias_joined_nick },
	{ ';',		alias_public_nick },
	{ '$',		alias_dollar },
	{ 'A',		alias_away },
	{ 'B',		alias_msg_body },
	{ 'C',		alias_channel },
	{ 'D',		alias_detected },
	{ 'E',		alias_idle },
	{ 'F',		alias_online },
/*	{ 'G' },
*/
	{ 'H', 		alias_current_numeric },
	{ 'I',		alias_invite },
	{ 'J',		alias_version_str },
	{ 'K',		alias_cmdchar },
	{ 'L',		alias_line },
	{ 'M',		alias_modes },
	{ 'N',		alias_nick },
	{ 'O',		alias_oper },
	{ 'P',		alias_chanop },
	{ 'Q',		alias_query_nick },
	{ 'R',		alias_server_version },
	{ 'S',		alias_server },
	{ 'T',		alias_target },
	{ 'U',		alias_buffer },
	{ 'V',		alias_version },
	{ 'W',		alias_currdir },
	{ 'X',		alias_show_userhost },
	{ 'Y',		alias_show_realname },
	{ 'Z',		alias_time },
	{ 0,	 	NULL }
};

	char	command_line[BIG_BUFFER_SIZE+1];

/* the 30 "standard" functions */
static	char	*function_channels 	_((char *));
static	char	*function_connect 	_((char *));
static	char	*function_curpos 	_((char *));
static	char	*function_decode 	_((unsigned char *));
static	char	*function_encode 	_((unsigned char *));
static	char	*function_index 	_((char *));
static	char	*function_ischannel 	_((char *));
static	char	*function_ischanop 	_((char *));
static	char	*function_left 		_((char *));
static	char	*function_listen 	_((char *));
static	char	*function_match 	_((char *));
static	char	*function_mid 		_((char *));
static	char	*function_pid 		_((char *));
static	char	*function_ppid 		_((char *));
static	char	*function_rand 		_((char *));
static	char	*function_right 	_((char *));
static	char	*function_rindex 	_((char *));
static	char	*function_rmatch 	_((char *));
static	char	*function_servers 	_((char *));
static	char	*function_srand 	_((char *));
static	char	*function_stime 	_((char *));
static	char	*function_strip 	_((char *));
static	char	*function_tdiff 	_((char *));
static	char	*function_time 		_((char *));
static	char	*function_tolower 	_((char *));
static	char	*function_toupper 	_((char *));
static	char	*function_userhost 	_((char *));
static	char	*function_winnum 	_((char *));
static	char	*function_winnam 	_((char *));
static	char	*function_word 		_((char *));


/* the 52 "extended" functions */
static	char *	function_after 		_((char *));
static	char *	function_afterw 	_((char *));
static	char *	function_ascii 		_((char *));
static	char *	function_before 	_((char *));
static	char *	function_beforew 	_((char *));
static	char *	function_center 	_((char *));
static	char *	function_chngw 		_((char *));
static	char *	function_chops 		_((char *));
static	char *	function_chr 		_((char *));
static	char *	function_close 		_((char *));
static	char *	function_common 	_((char *));
static	char *	function_convert 	_((char *));
static	char *	function_copattern 	_((char *));
static	char *	function_diff 		_((char *));
static	char *	function_eof 		_((char *));
static	char *	function_filter 	_((char *));
static	char *	function_fromw 		_((char *));
static	char *	function_insertw 	_((char *));
static	char *	function_iptoname 	_((char *));
static	char *	function_isalpha 	_((char *));
static	char *	function_isdigit 	_((char *));
static	char *	function_jot 		_((char *));
static	char *	function_key 		_((char *));
static	char *	function_leftw 		_((char *));
static	char *	function_midw 		_((char *));
static	char *	function_nametoip 	_((char *));
static	char *	function_nochops 	_((char *));
static	char *	function_notw 		_((char *));
static	char *	function_numonchannel 	_((char *));
static	char *	function_onchannel 	_((char *));
static	char *	function_open 		_((char *));
static	char *	function_pattern 	_((char *));
static	char *	function_read 		_((char *));
static	char *	function_remw 		_((char *));
static	char *	function_rename 	_((char *));
static	char *	function_restw 		_((char *));
static	char *	function_reverse 	_((char *));
static	char *	function_revw 		_((char *));
static	char *	function_rfilter 	_((char *));
static	char *	function_rightw 	_((char *));
static	char *	function_rmdir 		_((char *));
static	char *	function_rpattern 	_((char *));
static	char *	function_sar 		_((char *));
static	char *	function_server_version _((char *));
static	char *	function_split 		_((char *));
static	char *	function_splice 	_((char *));
static	char *	function_tow 		_((char *));
static	char *	function_translate 	_((char *));
static	char *	function_truncate 	_((char *));
static	char *	function_unlink 	_((char *));
static	char *	function_which 		_((char *));
static	char *	function_write 		_((char *));


typedef char *(bf) _((char *));
typedef struct
{
	char	*name;
	bf 	*func;
}	BuiltInFunctions;


/* 
 * This is the built-in function list.  This list *must* be sorted because
 * it is binary searched.   See the code for each function to see how it
 * is used.  Or see the help files.  Or see both.  Or look at the code
 * and see how it REALLY works, irregardless of the documentation >;-)
 */
static BuiltInFunctions	built_in_functions[] =
{
	{ "AFTER",              function_after 		},
	{ "AFTERW",             function_afterw 	},
	{ "ASCII",              function_ascii 		},
	{ "BEFORE",             function_before 	},
	{ "BEFOREW",            function_beforew 	},
	{ "CENTER",		function_center 	},
	{ "CHANUSERS",		function_onchannel 	},
	{ "CHNGW",              function_chngw 		},
	{ "CHOPS",              function_chops 		},
	{ "CHR",                function_chr 		},
	{ "CLOSE",		function_close 		},
	{ "COMMON",             function_common 	},
	{ "CONNECT",		function_connect 	},
	{ "CONVERT",		function_convert 	},
	{ "COPATTERN",          function_copattern 	},
	{ "CURPOS",		function_curpos 	},
	{ "DECODE",	  (bf *)function_decode 	},
	{ "DELARRAY",           function_delarray 	},
	{ "DELITEM",            function_delitem 	},
	{ "DIFF",               function_diff 		},
	{ "ENCODE",	  (bf *)function_encode 	},
	{ "EOF",		function_eof 		},
	{ "FILTER",             function_filter 	},
	{ "FINDITEM",           function_finditem 	},
	{ "FROMW",              function_fromw 		},
	{ "GETARRAYS",          function_getarrays 	},
	{ "GETITEM",            function_getitem 	},
	{ "GETMATCHES",         function_getmatches 	},
	{ "GETRMATCHES",        function_getrmatches 	},
	{ "IFINDFIRST",         function_ifindfirst 	},
	{ "IFINDITEM",          function_ifinditem 	},
	{ "IGETITEM",           function_igetitem 	},
	{ "INDEX",		function_index 		},
	{ "INDEXTOITEM",        function_indextoitem 	},
	{ "INSERTW",            function_insertw 	},
	{ "IPTONAME",		function_iptoname 	},
	{ "ISALPHA",		function_isalpha 	},
	{ "ISCHANNEL",		function_ischannel 	},
	{ "ISCHANOP",		function_ischanop 	},
	{ "ISDIGIT",		function_isdigit 	},
	{ "ITEMTOINDEX",        function_itemtoindex 	},
	{ "JOT",                function_jot 		},
	{ "KEY",                function_key 		},
	{ "LEFT",		function_left 		},
	{ "LEFTW",              function_leftw 		},
	{ "LISTEN",		function_listen 	},
	{ "MATCH",		function_match 		},
	{ "MATCHITEM",          function_matchitem 	},
	{ "MID",		function_mid 		},
	{ "MIDW",               function_midw 		},
	{ "MYCHANNELS",		function_channels 	},
	{ "MYSERVERS",		function_servers 	},
	{ "NAMETOIP",		function_nametoip 	},
	{ "NOCHOPS",            function_nochops 	},
	{ "NOTW",               function_notw 		},
	{ "NUMARRAYS",          function_numarrays 	},
	{ "NUMITEMS",           function_numitems 	},
	{ "NUMONCHANNEL",	function_numonchannel 	},
	{ "ONCHANNEL",          function_onchannel 	},
	{ "OPEN",		function_open 		},
	{ "PATTERN",            function_pattern 	},
	{ "PID",		function_pid 		},
	{ "POP",		function_pop 		},
	{ "PPID",		function_ppid 		},
	{ "PUSH",		function_push 		},
	{ "RAND",		function_rand 		},
	{ "READ",		function_read 		},
	{ "REMW",               function_remw 		},
	{ "RENAME",		function_rename 	},
	{ "RESTW",              function_restw 		},
	{ "REVERSE",            function_reverse 	},
	{ "REVW",               function_revw 		},
	{ "RFILTER",            function_rfilter 	},
	{ "RIGHT",		function_right 		},
	{ "RIGHTW",             function_rightw 	},
	{ "RINDEX",		function_rindex 	},
	{ "RMATCH",		function_rmatch 	},
	{ "RMATCHITEM",         function_rmatchitem 	},
	{ "RMDIR",		function_rmdir 		},
	{ "RPATTERN",           function_rpattern 	},
	{ "SAR",		function_sar 		},
	{ "SETITEM",            function_setitem 	},
	{ "SHIFT",		function_shift 		},
	{ "SPLICE",		function_splice 	},
	{ "SPLIT",		function_split 		},
	{ "SRAND",		function_srand 		},
	{ "STIME",		function_stime 		},
	{ "STRIP",		function_strip 		},
	{ "TDIFF",		function_tdiff 		},
	{ "TIME",		function_time 		},
	{ "TOLOWER",		function_tolower 	},
	{ "TOUPPER",		function_toupper 	},
	{ "TOW",                function_tow 		},
	{ "TR",			function_translate 	},
	{ "TRUNC",		function_truncate 	},
	{ "UNLINK",		function_unlink 	},
	{ "UNSHIFT",		function_unshift 	},
	{ "USERHOST",		function_userhost 	},
	{ "VERSION",		function_server_version },
	{ "WHICH",		function_which 		},
	{ "WINNAM",		function_winnam 	},
	{ "WINNUM",		function_winnum 	},
	{ "WORD",		function_word 		},
	{ "WRITE",		function_write 		},
	{ (char *) 0,		NULL }
};

#define	NUMBER_OF_ALIASES (sizeof(built_in_functions) / sizeof(BuiltInFunctions)) - 2

/* alias_illegals: characters that are illegal in alias names */
	char	alias_illegals[] = " #+-*/\\()={}[]<>!@$%^~`,?;:|'\"";

Alias	*alias_list[] =
{
	(Alias *) 0,
	(Alias *) 0
};

/* alias_string: the thing that gets replaced by the $"..." construct */
static	char	*alias_string = (char *) 0;

/* function_stack and function_stkptr - hold the return values from functions */
static	char	*function_stack[128] = { (char *) 0 };

static	int	eval_args;
static	int	function_stkptr = 0;


/* makes foo[one][two] look like tmp.one.two -- got it? */
#ifdef __STDC__
static void remove_brackets (char *name, char *args, int *arg_flag)
#else
static void remove_brackets (name, args, arg_flag)
char *name;
char *args;
int *arg_flag;
#endif
{
	char *str = NULL, *ptr, *right;
	char *result1;

	while ((ptr = strchr(name, '[')))
	{
		*ptr++ = 0;				/* name <- ptr */
		malloc_strcpy(&str, ptr);		/* str */
		right = str + 1;			/* str <- right */
		ptr = MatchingBracket(right, '[', ']'); /* str <- ptr */
		if (ptr)				/* str <- ptr */
			*ptr++ = 0;			/* str <- ptr */

		result1 = expand_alias(NULL, right, args, arg_flag, NULL);
		strcat(name, ".");
		strcat(name, result1);
		strcat(name, ptr);
		new_free(&result1);
		new_free(&str);
	}
}

/*
 * find_alias: looks up name in in alias list.  Returns the Alias entry if
 * found, or null if not found.   If unlink is set, the found entry is
 * removed from the list as well.  If match is null, only perfect matches
 * will return anything.  Otherwise, the number of matches will be returned. 
 */
#ifdef __STDC__
extern	Alias	*find_alias (Alias **list, char *name, int unlink, int *match, int unstub)
#else
extern	Alias	*find_alias (list, name, unlink, match, unstub)
	Alias	**list;
	char	*name;
	int	unlink;
	int	*match;
	int	unstub;
#endif
{
	Alias	*tmp,
		*last = (Alias *) 0;
	int	cmp,
		len;
	int (*cmp_func) ();

	if (match)
	{
		*match = 0;
		cmp_func = my_strnicmp;
	}
	else
		cmp_func = my_stricmp;

	if (name)
	{
		/* The function that is looking for the variable should not
		   be responsible for resolving any array variables.  That
		   should be and is our job.  So anyone can pass us a
		   variable name like 'foo[$one][two]' and we know just
		   what to do with it.  righto?  */
		int af;
		remove_brackets(name, NULL, &af);

		len = strlen(name);
		for (tmp = *list; tmp; tmp = tmp->next)
		{
			if ((cmp = cmp_func(name, tmp->name, len)) == 0)
			{
				static already_looking = 0;
				if (tmp->stub && unstub)
				{
					/* If already_looking is 1, then
					 * we are un-stubbing this alias
					 * because we are loading a file
					 * that presumably it must be in.
					 * So we dont load it again (duh).
					 */
					if (!already_looking)
					{
						already_looking = 1;
						load("LOAD", tmp->stub, empty_string);
						already_looking = 0;
					}
					/* it is very important for every
					 * caller of this function to assume
					 * that 'stub' is always NULL unless
					 * they are checking for non-NULLness.
					 * NO NEED to ever free or reassign
					 * the stub field anywhere but here!
					 */
					/* Whee.. recursion implies that
					 * the interesting parts of tmp have
					 * already been re-assigned by the 
					 * load, so we just clean up after
					 * ourselves.  We have to complete
					 * this call, becuase whoever called
					 * us is expecting a reply, so the load
					 * was just a temporary time-out while
					 * we feteched the interesting alias.
					 */
					new_free(&(tmp->stub));
					tmp->stub = NULL;

					/* I dont know if this will get
					   me in trouble or not, but im 
					   going to try it.  If we have
					   loaded the file, and this alias
					   is therefore de-stubbed, if
					   it wasnt assigned somewhere in
					   the file, then we should unlink
					   the stub. */
					if (!tmp->stuff)
						unlink = 1;
				}

				if (unlink)
				{
					if (last)
						last->next = tmp->next;
					else
						*list = tmp->next;
				}
				if (match)
				{
					(*match)++;
					if (strlen(tmp->name) == len)
					{
						*match = 0;
						return (tmp);
					}
				}
				else
					return (tmp);
			}
			else if (cmp < 0)
				break;
			last = tmp;
		}
	}
	if (match && (*match == 1))
		return (last);
	else
		return ((Alias *) 0);
}

/*
 * insert_alias: adds the given alias to the alias list.  The alias list is
 * alphabetized by name 
 */
#ifdef __STDC__
extern	void	insert_alias (Alias **list, Alias *alias)
#else
extern	void	insert_alias (list, alias)
	Alias	**list;
	Alias	*alias;
#endif
{
	Alias	*tmp,
		*last,
		*foo;

	last = (Alias *) 0;
	for (tmp = *list; tmp; tmp = tmp->next)
	{
		if (strcmp(alias->name, tmp->name) < 0)
			break;
		last = tmp;
	}
	if (last)
	{
		foo = last->next;
		last->next = alias;
		alias->next = foo;
	}
	else
	{
		alias->next = *list;
		*list = alias;
	}
}

/*
 * add_alias: given the alias name and "stuff" that makes up the alias,
 * add_alias() first checks to see if the alias name is already in use... if
 * so, the old alias is replaced with the new one.  If the alias is not
 * already in use, it is added. 
 */
#ifdef __STDC__
extern void	add_alias (int type, char *name, char *stuff)
#else
extern void	add_alias (type, name, stuff)
	int	type;
	char	*name,
		*stuff;
#endif
{
	Alias	*tmp;
	char	*ptr;

	upper(name);
	if (type == COMMAND_ALIAS)
		say("Alias	%s added", name);
	else
	{
		if (!strcmp(name, "FUNCTION_RETURN"))
		{
			if (function_stack[function_stkptr])
				new_free(&function_stack[function_stkptr]);
			malloc_strcpy(&function_stack[function_stkptr], stuff);
			return;
		}
		if ((ptr = sindex(name, alias_illegals)) != NULL)
		{
			yell("Assign names may not contain '%c'", *ptr);
			return;
		}
		say("Assign %s added", name);
	}

	/* dont forget -- if 'name' contains any [] pairs, find_alias will
	   evaluate and remove them for us... */
	/* We dont need to unstub an alias if we're just going to add a new one. */
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) 0, 0)) == (Alias *) 0)
	{
		tmp = (Alias *) new_malloc(sizeof(Alias));
		if (tmp == (Alias *) 0)
		{
			yell("Couldn't allocate memory for new alias!");
			return;
		}
		tmp->name = (char *) 0;
		tmp->stuff = (char *) 0;
		tmp->stub = (char *) 0;
	}
	malloc_strcpy(&(tmp->name), name);
	malloc_strcpy(&(tmp->stuff), stuff);
	tmp->mark = 0;
	tmp->global = loading_global;
	insert_alias(&(alias_list[type]), tmp);
}

/*
 * add_stub: You give it an alias name and a filename, and it generates
 * a stub for that alias.  If an alias is already in place for that name,
 * it is duly clobbered.  If it isnt already in place, this makes a new one.
 */
#ifdef __STDC__
static void	add_stub (int type, char *name, char *file)
#else
static void	add_stub (type, name, file)
	int	type;
	char	*name,
		*file;
#endif
{
	Alias	*tmp;
	char	*ptr;



	upper(name);
	if (type == COMMAND_ALIAS)
		say("Alias	%s stubbed to file %s", name, file);
	else
	{
		if (!strcmp(name, "FUNCTION_RETURN"))
		{
			yell("You cannot stub the FUNCTION_RETURN variable");
			return;
		}
		if ((ptr = sindex(name, alias_illegals)) != NULL)
		{
			yell("Assign names may not contain '%c'", *ptr);
			return;
		}
		say("Assign	%s stubbed to file %s", name, file);
	}
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) 0, 0)) == (Alias *) 0)
	{
		tmp = (Alias *)new_malloc(sizeof(Alias));
		if (tmp == (Alias *) 0)
		{
			yell("Couldn't allocate memory for new alias!");
			return;
		}
		tmp->name = (char *) 0;
		tmp->stuff = (char *) 0;
		tmp->stub = (char *) 0;
	}
	malloc_strcpy(&(tmp->name), name);
	malloc_strcpy(&(tmp->stub), file);
	new_free(&(tmp->stuff));
	tmp->stuff = NULL;
	tmp->mark = 0;
	tmp->global = loading_global;
	insert_alias(&(alias_list[type]), tmp);
}


/* canon_number: canonicalizes number to something relevant */
/* If FLOATING_POINT_MATH isnt set, it truncates it to an integer */
#ifdef __STDC__
static void canon_number (char *input)
#else
static void canon_number (input)
char *input;
#endif
{
	int end = strlen(input);

	if (end)
		end--;
	else
		return;		/* nothing to do */

	if (get_int_var(FLOATING_POINT_MATH_VAR))
	{
		/* remove any trailing zeros */
		while (input[end] == '0')
			end--;

		/* If we removed all the zeros and all that is
		   left is the decimal point, remove it too */
		if (input[end] == '.')
			end--;

		input[end+1] = 0;
	}
	else
	{
		char *dot = index(input, '.');
		if (dot)
			*dot = 0;
	}
}

/* word_count: returns the number of words in the given string */
#ifdef __STDC__
extern	int	word_count (char *str)
#else
extern	int	word_count (str)
	char	*str;
#endif
{
	char *foo = m_strdup(str);
	char *placeholder = foo;
	int count;

	count = 0;
	while (foo && *foo)
	{
		count++;
		next_arg(foo, &foo);
	}

	new_free(&placeholder);
	return count;
}

#ifdef __STDC__
static	char	*built_in_alias (char c)
#else
static	char	*built_in_alias (c)
	char	c;
#endif
{
	BuiltIns	*tmp;

	for (tmp = built_in;tmp->name;tmp++)
		if (c == tmp->name)
			return tmp->func();

	return m_strdup(empty_string);
}

/*
 * find_inline: This simply looks up the given str.  It first checks to see
 * if its a user variable and returns it if so.  If not, it checks to see if
 * it's an IRC variable and returns it if so.  If not, it checks to see if
 * its and environment variable and returns it if so.  If not, it returns
 * null.  It mallocs the returned string 
 */
#ifdef __STDC__
static	char	*find_inline (char *str)
#else
static	char	*find_inline (str)
	char	*str;
#endif
{
	Alias	*alias;
	char	*ret = NULL;
	char	*tmp;

	if ((alias = find_alias(&(alias_list[VAR_ALIAS]), str, 0, (int *) NULL, 1)) != NULL)
		return m_strdup(alias->stuff);
	if ((strlen(str) == 1) && (ret = built_in_alias(*str)))
		return(ret);
	if ((ret = make_string_var(str)) != NULL)
		return (ret);

	return m_strdup((tmp = getenv(str)) ? tmp : empty_string);
}

#ifdef __STDC__
static	char	*call_function (char *name, char *f_args, char *args, int *args_flag)
#else
static	char	*call_function (name, f_args, args, args_flag)
	char	*name,
		*f_args,
		*args;
	int	*args_flag;
#endif
{
	char	*tmp;
	char	*result = (char *) 0;
	char	*sub_buffer = (char *) 0;
	char	*debug_copy = (char *) 0;
	int min, max, pos, old_pos = -1, c;

	tmp = expand_alias((char *) 0, f_args, args, args_flag, NULL);
	if (get_int_var(DEBUG_VAR) & DEBUG_FUNCTIONS)
		malloc_strcpy(&debug_copy, tmp);

	upper(name);
	min = 0;
	max = NUMBER_OF_ALIASES + 1;
	while (1)
	{
		pos = (max + min) / 2;
		if (pos == old_pos)
		{
			pos = -1;
			break;
		}
		old_pos = pos;
		c = strcmp(name, built_in_functions[pos].name);
		if (c == 0)
			break;
		else if (c > 0)
			min = pos;
		else
			max = pos;
	}

	if (pos != -1)
		result = built_in_functions[pos].func(tmp);
	else
	{
		sub_buffer = new_malloc(strlen(name)+strlen(tmp)+2);
		strcpy(sub_buffer, name);
		strcat(sub_buffer, " ");
		strcat(sub_buffer, tmp);
		function_stack[++function_stkptr] = (char *) 0;
		parse_command(sub_buffer, 0, empty_string);
		new_free(&sub_buffer);
		eval_args=1;
		result = function_stack[function_stkptr];
		function_stack[function_stkptr] = (char *) 0;
		if (!result)
			malloc_strcpy(&result, empty_string);
		function_stkptr--;
	}
	if (debug_copy)
	{
		yell("Function %s(%s) returned %s", name, debug_copy, result);
		new_free(&debug_copy);
	}
	new_free(&tmp);
	return result;
}


/* Given a pointer to an operator, find the last operator in the string */
#ifdef __STDC__
static char	*lastop (char *ptr)
#else
static char	*lastop (ptr)
	char	*ptr;
#endif
{
	while (ptr[1] && index("!=<>&^|#+/%,-*", ptr[1]))
		ptr++;
	return ptr;
}

#define	NU_EXPR	0
#define	NU_ASSN	1
#define NU_TERT 2
#define NU_CONJ 3
#define NU_BITW 4 
#define NU_COMP 5 
#define NU_ADD  6 
#define NU_MULT 7 
#define NU_UNIT 8

#ifdef __STDC__
static	char	*next_unit (char *str, char *args, int *arg_flag, int stage)
#else
static	char	*next_unit (str, args, arg_flag, stage)
	char	*str,
		*args;
	int	*arg_flag,
		stage;
#endif
{
	char	*ptr,
		*ptr2,
		*right;
	int	got_sloshed = 0;
	char	*lastc;
	char	tmp[40];
	char	*result1 = (char *) 0,
		*result2 = (char *) 0;
	long	value1 = 0,
		value2,
		value3 = 0;
	double	dvalue1 = 0.0,
		dvalue2 = 0.0,
		dvalue3 = 0.0;
	char	op;
	int	display;
	char	*ArrayIndex,
		*EndIndex;

	while (isspace(*str))
		++str;
	if (!*str)
	{
		malloc_strcpy(&result1, empty_string);
		return result1;
	}
	lastc = str+strlen(str)-1;
	while (isspace(*lastc))
		*lastc-- = '\0';

	if (stage == NU_UNIT && *lastc == ')' && *str == '(')
	{
		str++, *lastc-- = '\0';
		return next_unit(str, args, arg_flag, NU_EXPR);
	}
	if (!*str)
	{
		malloc_strcpy(&result1, empty_string);
		return result1;
	}
	for (ptr = str; *ptr; ptr++)
	{
		if (got_sloshed)
		{
			got_sloshed = 0;
			continue;
		}

		switch(*ptr)
		{
		case '\\':
			got_sloshed = 1;
			continue;
		case '(':
			if (stage != NU_UNIT || ptr == str)
			{
				if (!(ptr2 = MatchingBracket(ptr+1, '(', ')')))
					ptr = ptr+strlen(ptr)-1;
				else
					ptr = ptr2;
				break;
			}
			*ptr++ = '\0';
			right = ptr;
			ptr = MatchingBracket(right, LEFT_PAREN, RIGHT_PAREN);
			if (ptr)
				*ptr++ = '\0';
			result1 = call_function(str, right, args, arg_flag);
			if (ptr && *ptr)
			{
				malloc_strcat(&result1, ptr);
				result2 = next_unit(result1, args, arg_flag, stage);
				new_free(&result1);
				result1 = result2;
			}
			return result1;
		case '[':
			if (stage != NU_UNIT)
			{
				if (!(ptr2 = MatchingBracket(ptr+1, '[', ']')))
					ptr = ptr+strlen(ptr)-1;
				else
					ptr = ptr2;
				break;
			}
			*ptr++ = '\0';
			right = ptr;
			ptr = MatchingBracket(right, LEFT_BRACKET, RIGHT_BRACKET);
			if (ptr)
				*ptr++ = '\0';
			result1 = expand_alias((char *) 0, right, args, arg_flag, NULL);
			if (*str)
			{
				result2 = new_malloc(strlen(str)+
						(result1?strlen(result1):0)+
						(ptr?strlen(ptr):0) + 2);
				strcpy(result2, str);
				strcat(result2, ".");
				strcat(result2, result1);
				new_free(&result1);
				if (ptr && *ptr)
				{
					strcat(result2, ptr);
					result1 = next_unit(result2, args, arg_flag, stage);
				}
				else
				{
					result1 = find_inline(result2);
					if (!result1)
						malloc_strcpy(&result1, empty_string);
				}
				new_free(&result2);
			}
			else if (ptr && *ptr)
			{
				malloc_strcat(&result1, ptr);
				result2 = next_unit(result1, args, arg_flag, stage);
				new_free(&result1);
				result1 = result2;
			}
			/* WARNING WARNING WARNING KLUDGE HERE */
			if (!*result1)
				malloc_strcpy(&result1, empty_string);
			return result1;
		case '-':
		case '+':
			if (*(ptr+1) == *(ptr))  /* index operator */
			{
				char *tptr;

				*ptr++ = '\0';
				if (ptr == str + 1)	/* Its a prefix */
					tptr = str + 2;
				else			/* Its a postfix */
					tptr = str;

				result1 = find_inline(tptr);
				if (!result1) 
					malloc_strcpy(&result1,"0");

				{           /* This isnt supposed to be
						attached to the if, so
						dont "fix" it. */
					long r;
					r = atol(result1);
					if (*ptr == '+')
						r++;
					else    r--;
					sprintf(tmp,"%ld",r);
					display = window_display;
					window_display = 0;
					add_alias(VAR_ALIAS,tptr,tmp);
					window_display = display;
				}
		/* A kludge?  Cheating?  Maybe.... */
				if (ptr == str + 1) 
				{
					*(ptr-1) = ' ';
					*ptr = ' ';
				} else            
				{
					if (*ptr == '+')
						*(ptr-1) = '-';
					else
						*(ptr-1) = '+';
					*ptr = '1';
				}
				ptr = str; 
				new_free(&result1);
				break;
			}
			if (ptr == str) /* It's unary..... do nothing */
				break;
			if (ptr[1] == '=') /* += operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					result2 = next_unit(ptr, args, arg_flag, stage);
					window_display = 0;
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}
					result1 = find_inline(varname);
					if (!result1 || !*result1)
						dvalue1 = 0;
					else
						dvalue1 = atof(result1);	
					dvalue2 = atof(result2);
					new_free(&result1);
					new_free(&result2);
					if (op == '-')
						dvalue3 = dvalue1 - dvalue2;
					else
						dvalue3 = dvalue1 + dvalue2;
					sprintf(tmp, "%f", dvalue3);
					canon_number(tmp);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, tmp);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result1);
					malloc_strcpy(&result2, tmp);
					return result2;
				}
			}
				
			if (stage != NU_ADD)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			dvalue1 = atof(result1);
			dvalue2 = atof(result2);
			new_free(&result1);
			new_free(&result2);
			if (op == '-')
				dvalue3 = dvalue1 - dvalue2;
			else
				dvalue3 = dvalue1 + dvalue2;
			sprintf(tmp, "%f", dvalue3);
			canon_number(tmp);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '/':
		case '*':
		case '%':
			if (ptr == str) /* It's unary..... do nothing */
				break;

			if (ptr[1] == '=') /* += operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					result2 = next_unit(ptr, args, arg_flag, stage);
					window_display = 0;
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}
					result1 = find_inline(varname);
					if (!result1 || !*result1)
						dvalue1 = 0.0;
					else
						dvalue1 = atof(result1);
					dvalue2 = atof(result2);
					new_free(&result1);
					new_free(&result2);
					if (op == '*')
						dvalue3 = dvalue1 * dvalue2;
					else if (op == '/')
					{
						if (dvalue2 == 0.0)
						{
							value3 = 0.0;
							yell("Division by zero!");
						}
						else
							dvalue3 = dvalue1 / dvalue2;
					}
					else if (op == '%')
					{
						if ((int)dvalue2 == 0)
						{
							dvalue3 = 0.0;
							yell("Mod by zero!");
						}
						else
							dvalue3 = (int)dvalue1 % (int)dvalue2;
					}
					sprintf(tmp, "%f", dvalue3);
					canon_number(tmp);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, tmp);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result1);
					malloc_strcpy(&result2, tmp);
					return result2;
				}
			}
				
			if (stage != NU_MULT)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			dvalue1 = atof(result1);
			dvalue2 = atof(result2);
			new_free(&result1);
			new_free(&result2);
			if (op == '/')
			{
				if (dvalue2)
					dvalue3 = dvalue1 / dvalue2;
				else
				{
					dvalue3 = 0.0;
					say("Division by zero");
				}
			}
			else if (op == '*')
				dvalue3 = dvalue1 * dvalue2;
			else
			{
				if ((int)dvalue2)
					dvalue3 = (int)dvalue1 % (int)dvalue2;
				else
				{
					value3 = 0.0;
					say("Mod by zero");
				}
			}
			sprintf(tmp, "%f", dvalue3);
			canon_number(tmp);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '#':
			if (ptr[1] == '#')
			{
				if (stage != NU_ADD)
				{
					ptr = lastop(ptr);
					break;
				}
				else
				{
					*ptr = '\0';
					ptr += 2;
					result1 = next_unit(str, args, arg_flag, stage);
					result2 = next_unit(ptr, args, arg_flag, stage);
					malloc_strcat(&result1, result2);
					new_free(&result2);
					return result1;
				}
			}
			if (ptr[1] == '=') /* #= operator */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					if (!result1 && !*result1)
						result1 = empty_string;
					malloc_strcat(&result1, result2);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			if (stage != NU_ADD || ptr[1] != '#')
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			malloc_strcat(&result1, result2);
			new_free(&result2);
			return result1;

	/* Reworked - Jeremy Nelson, Feb 1994
	 * & or && should both be supported, each with different
	 * stages, same with || and ^^.  Also, they should be
	 * short-circuit as well.
	 */
		case '&':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				value1 = atol(result1);
				if (value1)
				{
					result2 = next_unit(ptr, args, arg_flag, stage);
					value2 = atol(result2);
					value3 = value1 && value2;
				}
				else
					value3 = 0;
				new_free(&result1);
				new_free(&result2);
				tmp[0] = '0' + (value3?1:0);
				tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			else if (ptr[1] == '=')  /* &= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 &= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 & value2;
				sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
		case '|':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				value1 = atol(result1);
				if (!value1)
				{
					result2 = next_unit(ptr, args, arg_flag, stage);
					value2 = atol(result2);
					value3 = value1 || value2;
				}
				else
					value3 = 1;
				new_free(&result1);
				new_free(&result2);
				tmp[0] = '0' + (value3 ? 1 : 0);
				tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			else if (ptr[1] == '=')  /* |= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}


					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 |= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 | value2;
				sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
		case '^':
			if (ptr[0] == ptr[1])
			{
				if (stage != NU_CONJ)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 2;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				value1 = value1?1:0;
				value2 = value2?1:0;
				value3 = value1 ^ value2;
				new_free(&result1);
				new_free(&result2);
				tmp[0] = '0' + (value3 ? 1 : 0);
				tmp[1] = '\0';
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			else if (ptr[1] == '=')  /* ^= op */
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}	
				else
				{
					char *varname = (char *) 0;
					op = *ptr;
					*ptr = '\0';
					ptr+=2;

					varname = expand_alias((char *) 0, str,
						args, arg_flag, NULL);
					display = window_display;
					window_display = 0;
					result2 = next_unit(ptr, args, arg_flag, stage);
					lastc = varname + strlen(varname) - 1;
					while (lastc > varname && *lastc == ' ')
						*lastc-- = '\0';
					for (ptr = varname; *ptr == ' '; ptr++);
					while ((ArrayIndex = (char *) index(ptr, '[')) != NULL)
					{
					    *ArrayIndex++='.';
					    if ((EndIndex = MatchingBracket(ArrayIndex,
						LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
					    {
						*EndIndex++='\0';
						strcat(ptr, EndIndex);
					    }
					    else
						break;
					}

					result1 = find_inline(varname);
					value1 = atol(result1);
					value2 = atol(result2);
					value1 ^= value2;
					malloc_sprintf(&result1, "%d", value1);

					if (*ptr)
						add_alias(VAR_ALIAS, ptr, result1);
					else
						yell("Invalid assignment expression");
					window_display = display;
					new_free(&result2);
					return result1;
				}
			}
			else
			{
				if (stage != NU_BITW)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr = '\0';
				ptr += 1;
				result1 = next_unit(str, args, arg_flag, stage);
				result2 = next_unit(ptr, args, arg_flag, stage);
				value1 = atol(result1);
				value2 = atol(result2);
				new_free(&result1);
				new_free(&result2);
				value3 = value1 ^ value2;
				sprintf(tmp, "%ld",value3);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
		case '?':
		{
			/* I forgot to account for nested ?:s... */
			int nested_q = 0;

			if (stage != NU_TERT)
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);

			/* Look for the next '?' or ':' */
			ptr2 = sindex(ptr, "?:");

			/* We STOP when we hit a ':' and there isnt
			 * any ? to match it ('cept ourselves...)
			 *
			 * This allows us to continue even if we are
			 * are at nested_q == 0, if the next ?: is a ?.
			 */
			while (ptr2)
			{
				if (*ptr2 == ':')
				{
					if (!nested_q)
						break;
					nested_q--;
				}
				else if (*ptr2 == '?')
					nested_q++;
				ptr2 = sindex(ptr2+1, "?:");
			}
			
			/* Unbalanced :, or possibly missing */
			if (!ptr2)  /* ? but no :, ignore */
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr2++ = '\0';
			right = result1;
			value1 = parse_number(&right);
			if ((value1 == -1) && (*right == (char) 0))
				value1 = 0;
			if ( value1 == 0 )
				while (isspace(*right))
					*(right++) = '\0';
			if ( value1 || *right )
				result2 = next_unit(ptr, args, arg_flag, stage);
			else
				result2 = next_unit(ptr2, args, arg_flag, stage);
			*(ptr2-1) = ':';
			new_free(&result1);
			return result2;
		}
		case '=':
			if (ptr[1] != '=')
			{
				if (stage != NU_ASSN)
				{
					ptr = lastop(ptr);
					break;
				}
				*ptr++ = '\0';
				result1 = expand_alias((char *) 0, str,
					args, arg_flag, NULL);
				result2 = next_unit(ptr, args, arg_flag, stage);
				display = window_display;
				window_display = 0;
				lastc = result1 + strlen(result1) - 1;
				while (lastc > result1 && *lastc == ' ')
					*lastc-- = '\0';
				for (ptr = result1; *ptr == ' '; ptr++);
				while ((ArrayIndex = (char *) index(ptr, '['))
						!= NULL)
				{
				    *ArrayIndex++='.';
				    if ((EndIndex = MatchingBracket(ArrayIndex,
					LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
				    {
					*EndIndex++='\0';
					strcat(ptr, EndIndex);
				    }
				    else
					break;
				}
				if (*ptr)
					add_alias(VAR_ALIAS, ptr, result2);
				else
					yell("Invalid assignment expression");
				window_display = display;
				new_free(&result1);
				return result2;
			}
			if (stage != NU_COMP)
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (!my_stricmp(result1, result2))
				malloc_strcpy(&result1, "1");
			else
				malloc_strcpy(&result1, "0");
			new_free(&result2);
			return result1;
		case '>':
		case '<':
			if (ptr[1] == ptr[0])  /* << or >> op */
			{
				/* Not supported yet */
				yell("<< or >> not supported yet");
				break;
			}
			else if (stage != NU_COMP)
			{
				ptr = lastop(ptr);
				break;
			}
			op = *ptr;
			if (ptr[1] == '=')
				value3 = 1, *ptr++ = '\0';
			else
				value3 = 0;
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (isdigit(*result1) && isdigit(*result2))
			{
				value1 = atol(result1);
				value2 = atol(result2);
				value1 = (value1 == value2) ? 0 : ((value1 <
					value2) ? -1 : 1);
			}
			else
				value1 = my_stricmp(result1, result2);
			if (value1)
			{
				value2 = (value1 > 0) ? 1 : 0;
				if (op == '<')
					value2 = 1 - value2;
			}
			else
				value2 = value3;
			new_free(&result2);
			sprintf(tmp, "%ld", value2);
			malloc_strcpy(&result1, tmp);
			return result1;
		case '~':
			if (ptr == str)
			{
				if (stage != NU_UNIT)
					break;
				result1 = next_unit(str+1, args, arg_flag, stage);
				if (isdigit(*result1))
				{
					value1 = atol(result1);
					value2 = ~value1;
				}
				else
					value2 = 0;
				sprintf(tmp, "%ld", value2);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			else
			{
				ptr = lastop(ptr);
				break;
			}
		case '!':
			if (ptr == str)
			{
				if (stage != NU_UNIT)
					break;
				result1 = next_unit(str+1, args, arg_flag, stage);

				/* digits can start with -'s, cant they? */
				if (isdigit(*result1) || 
				   ( (*result1 == '-') && isdigit(result1[1]) ))
				{
					value1 = atol(result1);
					value2 = value1 ? 0 : 1;
				}
				else
				{
					value2 = ((*result1)?0:1);
				}
				sprintf(tmp, "%ld", value2);
				malloc_strcpy(&result1, tmp);
				return result1;
			}
			if (stage != NU_COMP || ptr[1] != '=')
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr = '\0';
			ptr += 2;
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			if (!my_stricmp(result1, result2))
				malloc_strcpy(&result1, "0");
			else
				malloc_strcpy(&result1, "1");
			new_free(&result2);
			return result1;
		case ',': 
			/*
			 * this utterly kludge code is needed (?) to get
			 * around bugs introduced from hop's patches to
			 * alias.c.  the $, variable stopped working
			 * because of this.  -mrg, july 94.
			 */
			if (ptr == str || (ptr > str && ptr[-1] == '$'))
				break;
			if (stage != NU_EXPR)
			{
				ptr = lastop(ptr);
				break;
			}
			*ptr++ = '\0';
			result1 = next_unit(str, args, arg_flag, stage);
			result2 = next_unit(ptr, args, arg_flag, stage);
			new_free(&result1);
			return result2;
		}
	}
	if (stage != NU_UNIT)
		return next_unit(str, args, arg_flag, stage + 1);
	if (isdigit(*str) || *str == '+' || *str == '-')
		malloc_strcpy(&result1, str);
	else
	{
		if (*str == '#' || *str=='@')
			op = *str++;
		else
			op = '\0';
		result1 = find_inline(str);
		if (!result1)
			malloc_strcpy(&result1, empty_string);
		if (op)
		{
			if (op == '#')
				value1 = word_count(result1);
			else if (op == '@')
				value1 = strlen(result1);
			sprintf(tmp, "%ld", value1);
			malloc_strcpy(&result1, tmp);
		}
	}
	return result1;
}

/*
 * parse_inline:  This evaluates user-variable expression.  I'll talk more
 * about this at some future date. The ^ function and some fixes by
 * troy@cbme.unsw.EDU.AU (Troy Rollo) 
 */
extern
#ifdef __STDC__
char	*parse_inline(char *str, char *args, int *args_flag)
#else
char	*parse_inline(str, args, args_flag)
	char	*str;
	char	*args;
	int	*args_flag;
#endif
{
	return next_unit(str, args, args_flag, NU_EXPR);
}

/*
 * parse_number: returns the next number found in a string and moves the
 * string pointer beyond that point	in the string.  Here's some examples: 
 *
 * "123harhar"  returns 123 and str as "harhar" 
 *
 * while: 
 *
 * "hoohar"     returns -1  and str as "hoohar" 
 */
#ifdef __STDC__
extern	int	parse_number(char **str)
#else
extern	int	parse_number(str)
	char	**str;
#endif
{
	int	ret;
	char	*ptr;

	ptr = *str;
	if (isdigit(*ptr))
	{
		ret = atoi(ptr);
		for (; isdigit(*ptr); ptr++);
		*str = ptr;
	}
	else
		ret = -1;
	return (ret);
}

#ifdef __STDC__
static void	do_alias_string (char unused, char *not_used)
#else
static void do_alias_string (unused, not_used)
char unused, *not_used;
#endif
{
	malloc_strcpy(&alias_string, get_input());
}

/*
 * expander_addition: This handles string width formatting for irc variables
 * when [] is specified.  
 */
#ifdef __STDC__
static	void	expander_addition(char *buff, char *add, int length, char *quote_em)
#else
static	void	expander_addition(buff, add, length, quote_em)
	char	*buff,
		*add;
	int	length;
	char	*quote_em;
#endif
{
	char	format[40],
		*ptr;
	char	buffer[BIG_BUFFER_SIZE];

	if (length)
	{
		sprintf(format, "%%%d.%ds", -length, (length < 0 ? -length : length));
		sprintf(buffer, format, add);
		add = buffer;
	}
	if (quote_em)
	{
		ptr = double_quote(add, quote_em);
		strmcat(buff, ptr, BIG_BUFFER_SIZE);
		new_free(&ptr);
	}
	else if (buff)
		strmcat(buff, add, BIG_BUFFER_SIZE);
}

/* MatchingBracketBackwards returns the matching bracket of the given type */
extern
#ifdef __STDC__
char	*MatchingBracketBackwards(char *string, char *beg, char left, char right)
#else
char	*MatchingBracketBackwards(string, beg, left, right)
	char	*string;
	char	*beg;
	char	left,
		right;
#endif
{
	int	bracket_count = 1;

	while ((string > beg) && bracket_count)
	{
		if (*string == left)
			bracket_count++;
		else if (*string == right)
		{
			if (!--bracket_count)
				return string;
		}
		else if (string[-1] == '\\')
			string--;
		string--;
	}
	return (char *) 0;
}

/* MatchingBracket returns the next unescaped bracket of the given type */
extern
#ifdef __STDC__
char	*MatchingBracket(char *string, char left, char right)
#else
char	*MatchingBracket(string, left, right)
	char	*string;
	char	left,
		right;
#endif
{
	int	bracket_count = 1;

	while (*string && bracket_count)
	{
		if (*string == left)
			bracket_count++;
		else if (*string == right)
		{
			if (!--bracket_count)
				return string;
		}
		else if (*string == '\\' && string[1])
			string++;
		string++;
	}
	return (char *) 0;
}


/*
 * alias_special_char: Here we determin what to do with the character after
 * the $ in a line of text. The special characters are described more fulling
 * in the help/ALIAS file.  But they are all handled here. Paremeters are the
 * name of the alias (if applicable) to prevent deadly recursion, a
 * destination buffer (of size BIG_BUFFER_SIZE) to which things are appended,
 * a ptr to the string (the first character of which is the special
 * character, the args to the alias, and a character indication what
 * characters in the string should be quoted with a backslash.  It returns a
 * pointer to the character right after the converted alias.

 The args_flag is set to 1 if any of the $n, $n-, $n-m, $-m, $*, or $() is used
 in the alias.  Otherwise it is left unchanged.
 */
#ifdef __STDC__
static	char	*alias_special_char(char *name, char *buffer, char *ptr, char *args, char *quote_em, int *args_flag)
#else
static	char	*alias_special_char(name, buffer, ptr, args, quote_em,args_flag)
	char	*name;
	char	*buffer;
	char	*ptr;
	char	*args;
	char	*quote_em;
	int	*args_flag;
#endif
{
	char	*tmp,
		c;
	int	upper,
	lower,
	length;

	length = 0;
	if ((c = *ptr) == LEFT_BRACKET)
	{
		ptr++;
		if ((tmp = (char *) index(ptr, RIGHT_BRACKET)) != NULL)
		{
			*(tmp++) = (char) 0;
			length = atoi(ptr);
			ptr = tmp;
			c = *ptr;
		}
		else
		{
			say("Missing %c", RIGHT_BRACKET);
			return (ptr);
		}
	}
	tmp = ptr+1;
	switch (c)
	{
		case LEFT_PAREN:
		{
			char	sub_buffer[BIG_BUFFER_SIZE+1];

			if ((ptr = MatchingBracket(tmp, LEFT_PAREN, RIGHT_PAREN)) || 
			    (ptr = (char *) index(tmp, RIGHT_PAREN)))
				*(ptr++) = (char) 0;
			tmp = expand_alias((char *) 0, tmp, args, args_flag, NULL);
			*sub_buffer = (char) 0;
			alias_special_char((char *) 0, sub_buffer, tmp,
				args, quote_em,args_flag);
			expander_addition(buffer, sub_buffer, length, quote_em);
			new_free(&tmp);
			*args_flag = 1;
			return (ptr);
		}
		case '!':
		{
			if ((ptr = (char *) index(tmp, '!')) != NULL)
				*(ptr++) = (char) 0;
			if ((tmp = do_history(tmp, empty_string)) != NULL)
			{
				expander_addition(buffer, tmp, length, quote_em);
				new_free(&tmp);
			}
			return (ptr);
		}
		case LEFT_BRACE:
		{
			/* BLAH. This didnt allow for nesting before.  How lame. */
			ptr = MatchingBracket(tmp, LEFT_BRACE, RIGHT_BRACE);
			if (!ptr)
			{
				yell("Unmatched %c", LEFT_BRACE);
				break;
			}
			*(ptr++) = (char) 0;
			if ((tmp = parse_inline(tmp, args, args_flag)) != NULL)
			{
				expander_addition(buffer, tmp, length, quote_em);
				new_free(&tmp);
			}
			return (ptr);
		}
		case DOUBLE_QUOTE:
		{
			if ((ptr = (char *) index(tmp, DOUBLE_QUOTE)) != NULL)
				*(ptr++) = (char) 0;
			alias_string = (char *) 0;
			get_line(tmp, 0, do_alias_string);
			expander_addition(buffer, alias_string, length, quote_em);
			new_free(&alias_string);
			return (ptr);
		}
		case '*':
		{
			expander_addition(buffer, args, length, quote_em);
			*args_flag = 1;
			return (ptr + 1);
		}
		default:
		{
			if (isdigit(c) || (c == '-') || c == '~')
			{
				*args_flag = 1;
				if (c == '~')
				{
					/* double check to make sure $~ still works */
					lower = upper = EOS;
					ptr++;
				}
				else if (c == '-')
				{
					lower = 0;
					ptr++;
					upper = parse_number(&ptr);
					if (upper == -1)
						return empty_string; /* error */
				}
				else
				{
					lower = parse_number(&ptr);
					if (*ptr == '-')
					{
						ptr++;
						upper = parse_number(&ptr);
						if (upper == -1)
							upper = EOS;
					}
					else
						upper = lower;
				}
				/* this is a memory leak */
				expander_addition(buffer, extract2(args, lower, upper), length, quote_em);
				return (ptr ? ptr : empty_string);
			}
			else
			{
				char	*rest,
					c = (char) 0;

			/*
			 * Why use ptr+1?  Cause try to maintain backward compatability
			 * can be a pain in the butt.  Basically, we don't want any of
			 * the illegal characters in the alias, except that things like
			 * $* and $, were around first, so they must remain legal.  So
			 * we skip the first char after the $.  Does this make sense?
			 */
				/* special case for $ */
				if (*ptr == '$')
				{
					rest = ptr+1;
					c = *rest;
					*rest = (char) 0;
				}
				else if ((rest = sindex(ptr+1, alias_illegals)) != NULL)
				{
					if (isalpha(*ptr) || *ptr == '_')
						while ((*rest == LEFT_BRACKET ||
						    *rest == LEFT_PAREN) &&
						    (tmp = MatchingBracket(rest+1,
						    *rest, (*rest == LEFT_BRACKET) ?
						    RIGHT_BRACKET: RIGHT_PAREN)))
							rest = tmp + 1;
					c = *rest;
					*rest = (char) 0;
				}
				if ((tmp = parse_inline(ptr, args, args_flag)) != NULL)
				{
					expander_addition(buffer, tmp, length, quote_em);
					new_free(&tmp);
				}
				if (rest)
					*rest = c;
				return(rest);
			}
		}
	}
	return NULL;
}


/*
 * expand_alias: Expands inline variables in the given string and returns the
 * expanded string in a new string which is malloced by expand_alias(). 
 *
 * Also unescapes anything that was quoted with a backslash
 *
 * Behaviour is modified by the following:
 *	Anything between brackets (...) {...} is left unmodified.
 *	If more_text is supplied, the text is broken up at
 *		semi-colons and returned one at a time. The unprocessed
 *		portion is written back into more_text.
 *	Backslash escapes are unescaped.
 */
extern
#ifdef __STDC__
char	*expand_alias(char *name, char *string, char *args, int *args_flag, char **more_text)
#else
char	*expand_alias(name, string, args,args_flag, more_text)
	char	*name,
		*string,
		*args;
	int	*args_flag;
	char	**more_text;
#endif
{
	char	buffer[BIG_BUFFER_SIZE + 1],
		*ptr,
		*stuff = (char *) 0,
		*free_stuff;
	char	*quote_em,
		*quote_str = (char *) 0;
	char	ch;
	int	quote_cnt = 0;
	int	is_quote = 0;
	char *	(*str_cat) _((char *, char *, int));

	/* This is a kludge i put in here until i can figure
	   out why this is recursive */
	/* this *might* be a memory leak... just keep an eye on it */
	if (!string || !*string)
		return m_strdup(empty_string);

	if (*string == '@' && more_text)
	{
		str_cat = strmcat;
		*args_flag = 1; /* Stop the @ command from auto appending */
	}
	else
		str_cat = strmcat_ue;
	malloc_strcpy(&stuff, string);
	free_stuff = stuff;
	*buffer = (char) 0;
	eval_args = 1;
	ptr = stuff;
	if (more_text)
		*more_text = NULL;
	while (ptr && *ptr)
	{
		if (is_quote)
		{
			is_quote = 0;
			++ptr;
			continue;
		}
		switch(*ptr)
		{
		case '$':
	/*
	 * The test here ensures that if we are in the expression
	 * evaluation command, we don't expand $. In this case we
	 * are only coming here to do command separation at ';'s.
	 * If more_text is not defined, and the first character is
	 * '@', we have come here from [] in an expression.
	 */
			if (more_text && *string == '@')
			{
				ptr++;
				break;
			}
			*(ptr++) = (char) 0;
			(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
			while (*ptr == '^')
			{
				ptr++;
				if (quote_str)
					quote_str = (char *) new_realloc((char *)quote_str, sizeof(char) * (quote_cnt + 2));
				else
					quote_str = (char *) new_malloc(sizeof(char) * (quote_cnt + 2));
				quote_str[quote_cnt++] = *(ptr++);
				quote_str[quote_cnt] = (char) 0;
			}
			quote_em = quote_str;
			/* MEMORY LEAK */
			stuff = alias_special_char(name, buffer, ptr, args, quote_em, args_flag);
			if (stuff)
				new_free(&quote_str);
			quote_cnt = 0;
			ptr = stuff;
			break;
		case ';':
			if (!more_text)
			{
				ptr++;
				break;
			}
			*more_text = string + (ptr - free_stuff) +1;
			*ptr = '\0'; /* To terminate the loop */
			break;
		case LEFT_PAREN:
		case LEFT_BRACE:
			ch = *ptr;
			*ptr = '\0';
			(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
			stuff = ptr;
			*args_flag = 1;
			if (!(ptr = MatchingBracket(stuff + 1, ch,
					(ch == LEFT_PAREN) ?
					RIGHT_PAREN : RIGHT_BRACE)))
			{
				yell("Unmatched %c", ch);
				ptr = stuff + strlen(stuff+1)+1;
			}
			else
				ptr++;
			*stuff = ch;
			ch = *ptr;
			*ptr = '\0';
			strmcat(buffer, stuff, BIG_BUFFER_SIZE);
			stuff = ptr;
			*ptr = ch;
			break;
		case '\\':
			is_quote = 1;
			ptr++;
			break;
		default:
			ptr++;
			break;
		}
	}
	if (stuff)
		(*str_cat)(buffer, stuff, BIG_BUFFER_SIZE);
	ptr = (char *) 0;
	new_free(&free_stuff);
	malloc_strcpy(&ptr, buffer);
	if (get_int_var(DEBUG_VAR) & DEBUG_EXPANSIONS)
		yell("Expanded [%s] to [%s]",
			string, ptr);
	return (ptr);
}

/*
 * get_alias: returns the alias matching 'name' as the function value. 'args'
 * are expanded as needed, etc.  If no matching alias is found, null is
 * returned, cnt is 0, and full_name is null.  If one matching alias is
 * found, it is retuned, with cnt set to 1 and full_name set to the full name
 * of the alias.  If more than 1 match are found, null is returned, cnt is
 * set to the number of matches, and fullname is null. NOTE: get_alias()
 * mallocs the space for the full_name, but returns the actual value of the
 * alias if found! 
 */
extern
#ifdef __STDC__
char	*get_alias(int type, char *name, int *cnt, char **full_name)
#else
char	*get_alias(type, name, cnt, full_name)
	int	type;
	char	*name,
		**full_name;
	int	*cnt;
#endif
{
	Alias	*tmp;

	*full_name = (char *) 0;
	if ((name == (char *) 0) || (*name == (char) 0))
	{
		*cnt = 0;
		return ((char *) 0);
	}
	if ((tmp = find_alias(&(alias_list[type]), name, 0, cnt, 1)) != NULL)
	{
		if (*cnt < 2)
		{
			malloc_strcpy(full_name, tmp->name);
			return (tmp->stuff);
		}
	}
	return ((char *) 0);
}

/*
 * match_alias: this returns a list of alias names that match the given name.
 * This is used for command completion etc.  Note that the returned array is
 * malloced in this routine.  Returns null if no matches are found 
 */
extern
#ifdef __STDC__
char	**match_alias(char *name, int *cnt, int type)
#else
char	**match_alias(name, cnt, type)
	char	*name;
	int	*cnt;
	int	type;
#endif
{
	Alias	*tmp;
	char	**matches = (char **) 0;
	int	matches_size = 5;
	int	len;
	char	*last_match = (char *) 0;
	char	*dot;

	len = strlen(name);
	*cnt = 0;
	matches = (char	**) new_malloc(sizeof(char *) * matches_size);
	for (tmp = alias_list[type]; tmp; tmp = tmp->next)
	{
		if (strncmp(name, tmp->name, len) == 0)
		{
			if ((dot = (char *) index(tmp->name+len, '.')) != NULL)
			{
				if (type == COMMAND_ALIAS)
					continue;
				else
				{
					*dot = '\0';
					if (last_match && !strcmp(last_match,
							tmp->name))
					{
						*dot = '.';
						continue;
					}
				}
			}
			matches[*cnt] = (char *) 0;
			malloc_strcpy(&(matches[*cnt]), tmp->name);
			last_match = matches[*cnt];
			if (dot)
				*dot = '.';
			if (++(*cnt) == matches_size)
			{
				matches_size += 5;
				matches = (char	**) new_realloc((char *)matches,
					sizeof(char *) * matches_size);
			}
		}
		else if (*cnt)
			break;
	}
	if (*cnt)
	{
		matches = (char	**) new_realloc((char *)matches, sizeof(char *) * (*cnt + 1));
		matches[*cnt] = (char *) 0;
	}
	else
		new_free((char **)&matches);
	return (matches);
}

/* delete_alias: The alias name is removed from the alias list. */
extern
#ifdef __STDC__
void delete_alias(int type, char *name)
#else
void delete_alias(type, name)
	int	type;
	char	*name;
#endif
{
	Alias	*tmp;

	upper(name);

	/* we dont need to unstub it if we're just going to delete it */
	if ((tmp = find_alias(&(alias_list[type]), name, 1, (int *) NULL, 0))
			!= NULL)
	{
		new_free(&(tmp->name));
		new_free(&(tmp->stuff));
		new_free((char **)&tmp);
		if (type == COMMAND_ALIAS)
			say("Alias	%s removed", name);
		else
			say("Assign %s removed", name);
	}
	else
		say("No such alias: %s", name);
}

/*
 * list_aliases: Lists all aliases matching 'name'.  If name is null, all
 * aliases are listed 
 */
#ifdef __STDC__
static void list_aliases(int type, char *name)
#else
static void list_aliases(type, name)
	int	type;
	char	*name;
#endif
{
	Alias	*tmp;
	int	len;
	int	DotLoc,
		LastDotLoc = 0;
	char	*LastStructName = NULL;
	char	*s;

	if (type == COMMAND_ALIAS)
		say("Aliases:");
	else
		say("Assigns:");

	if (name)
	{
		upper(name);
		len = strlen(name);
	}
	else
		len = 0;
	for (tmp = alias_list[type]; tmp; tmp = tmp->next)
	{
		if (!name || !strncmp(tmp->name, name, len))
		{
			s = index(tmp->name + len, '.');
			if (!s)
			{
				if (tmp->stub)
					say("\t%s STUBBED TO %s", tmp->name, tmp->stub);
				else
					say("\t%s\t%s", tmp->name, tmp->stuff);
			}
			else
			{
				DotLoc = s - tmp->name;
				if (!LastStructName || (DotLoc != LastDotLoc) || strncmp(tmp->name, LastStructName, DotLoc))
				{
					say("\t%*.*s\t<Structure>", DotLoc, DotLoc, tmp->name);
					LastStructName = tmp->name;
					LastDotLoc = DotLoc;
				}
			}
		}
	}
}

/*
 * mark_alias: sets the mark field of the given alias to 'flag', and returns
 * the previous value of the mark.  If the name is not found, -1 is returned.
 * This is used to prevent recursive aliases by marking and unmarking
 * aliases, and not reusing an alias that has previously been marked.  I'll
 * explain later 
 */
static
#ifdef __STDC__
int mark_alias(char *name, int flag)
#else
int mark_alias(name, flag)
	char	*name;
	int	flag;
#endif
{
	int	old_mark;
	Alias	*tmp;
	int	match;

	if ((tmp = find_alias(&(alias_list[COMMAND_ALIAS]), name, 0, &match, 1)) != NULL)
	{
		if (match < 2)
		{
			old_mark = tmp->mark;
		/* New handling of recursion */
			if (flag)
			{
				int	i;
				/* Count recursion */

				tmp->mark = tmp->mark + flag;
				if ((i = get_int_var(MAX_RECURSIONS_VAR)) > 1)
				{
					if (tmp->mark > i)
					{
						tmp->mark = 0;
						return(1); /* MAX exceeded. */
					}
					else return(0);
				/* In recursion but it's ok */
				}
				else
				{
					if (tmp->mark > 1)
					{
						tmp->mark = 0;
						return(1);
				/* max of 1 here.. exceeded */
					}
					else return(0);
				/* In recursion but it's ok */
				}
			}
			else
		/* Not in recursion at all */
			{
				tmp->mark = 0;
				return(old_mark);
			/* This one gets ignored anyway */
			}
		}
	}
	return (-1);
}

/*
 * execute_alias: After an alias has been identified and expanded, it is sent
 * here for proper execution.  This routine mainly prevents recursive
 * aliasing.  The name is the full name of the alias, and the alias is
 * already expanded alias (both of these parameters are returned by
 * get_alias()) 
 */
#ifdef __STDC__
extern void execute_alias(char *alias_name, char *alias, char *args)
#else
extern void execute_alias(alias_name, alias, args)
	char	*alias_name,
		*alias,
		*args;
#endif
{
	if (mark_alias(alias_name, 1))
		say("Maximum recursion count exceeded in: %s", alias_name);
	else
	{
		parse_line(alias_name, alias, args, 0,1);
		mark_alias(alias_name, 0);
	}
}

/*
 * save_aliases: This will write all of the aliases to the FILE pointer fp in
 * such a way that they can be read back in using LOAD or the -l switch 
 */
extern
#ifdef __STDC__
void save_aliases(FILE *fp, int do_all)
#else
void save_aliases(fp, do_all)
	FILE	*fp;
	int	do_all;
#endif
{
	Alias	*tmp;

	for (tmp = alias_list[VAR_ALIAS]; tmp; tmp = tmp->next)
	{
		if (!tmp->global || do_all)
		{
			if (tmp->stub)
				fprintf(fp, "STUB ASSIGN %s %s\n", tmp->name, tmp->stub);
			else
				fprintf(fp, "ASSIGN %s %s\n", tmp->name, tmp->stuff);
		}
	}
	for (tmp = alias_list[COMMAND_ALIAS]; tmp; tmp = tmp->next)
	{
		if (!tmp->global || do_all)
		{
			if (tmp->stub)
				fprintf(fp, "STUB ALIAS %s %s\n", tmp->name, tmp->stub);
			else
				fprintf(fp, "ALIAS %s %s\n", tmp->name, tmp->stuff);
		}
	}	
}


#ifdef __STDC__
extern  void	flush_aliases(char *command,char *args, char *subargs)
#else
extern  void	flush_aliases(command,args,subargs)
	char 	*command,
		*args;
	char	*subargs;
#endif
{
	char 	*blah;

	if (!args || !*args)
	{
		flush_aliases2();
		flush_variables();
		flush_on_hooks();
		return;
	}
	while ((blah = next_arg(args, &args)))
	{
		if (!my_strnicmp(blah,"VAR",1))
			flush_variables();
		else if (!my_strnicmp(blah,"ALIAS",3))
			flush_aliases2();
		else if (!my_strnicmp(blah,"ON",1))
			flush_on_hooks();
		else if (!my_strnicmp(blah,"ALL",3))
		{
			flush_aliases2();
			flush_variables();
			flush_on_hooks();
		}
	}
	yell("*** I hope you didnt have anything valuable in there...");
}

static void flush_aliases2 _((void))
{
	Alias	*tmp;
	Alias   *ntmp;

	for (tmp = alias_list[COMMAND_ALIAS]; tmp;)
	{
		/* we dont need to unstub if we're just going to flush it. */
		tmp = find_alias(&alias_list[COMMAND_ALIAS],tmp->name,1,NULL, 0);
		new_free(&tmp->stuff);
		new_free(&tmp->name);
		ntmp = tmp->next;
		new_free((char **)&tmp);
		tmp = ntmp;
	}
	alias_list[COMMAND_ALIAS] = (Alias *) 0;
}

static void flush_variables _((void))
{
	Alias	*tmp;
	Alias   *ntmp;

	for (tmp = alias_list[VAR_ALIAS]; tmp;)
	{
		/* we dont need to unstub if we're just going to flush it. */
		tmp = find_alias(&alias_list[VAR_ALIAS],tmp->name,1,NULL, 0);
		new_free(&tmp->stuff);
		new_free(&tmp->name);
		ntmp = tmp->next;
		new_free((char **)&tmp);
		tmp = ntmp;
	}
	alias_list[VAR_ALIAS] = (Alias *) 0;
}



/*
 * alias: the /ALIAS command.  Calls the correct alias function depending on
 * the args 
 */
extern
#ifdef __STDC__
void	alias(char *command, char *args, char *subargs)
#else
void	alias(command, args, subargs)
	char	*command,
		*args;
	char	*subargs;
#endif
{
	char	*name,
		*rest;
	int	type;
	char	*ArrayIndex;
	char	*EndIndex;

	type = *command - 48;	/*
				 * A trick!  Yes, well, what the hell.  Note
				 * the the command part of ALIAS is "0" and
				 * the command part of ASSIGN is "1" in the
				 * command array list
				 */
	if ((name = next_arg(args, &rest)) != NULL)
	{
		while ((ArrayIndex = (char *) index(name, '[')) != NULL)
		{
			*ArrayIndex++ = '.';
			if ((EndIndex = MatchingBracket(ArrayIndex, LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
			{
				*EndIndex++ = '\0';
				strcat(name, EndIndex);
			}
			else
				break;
		}
		if (*rest)
		{
			if (*rest == LEFT_BRACE)
			{
				char	*ptr = MatchingBracket(++rest, LEFT_BRACE, RIGHT_BRACE);
				if (!ptr)
					say("Unmatched brace in ALIAS or ASSIGN");
				else if (ptr[1])
					say("Junk after closing brace in ALIAS or ASSIGN");
				else
				{
					*ptr = '\0';
					add_alias(type, name, rest);
				}
			}
			else
				add_alias(type, name, rest);
		}
		else
		{
			if (*name == '-')
			{
				if (*(name + 1))
					delete_alias(type, name + 1);
				else
					say("You must specify an alias to be removed");
			}
			else
				list_aliases(type, name);
		}
	}
	else
		list_aliases(type, (char *) 0);
}

/* stub command -- makes a stub for an alias/assign */
extern
#ifdef __STDC__
void stubcmd (char *command, char *args, char *subargs)
#else
void stubcmd (command, args, subargs)
char *command;
char *args;
char *subargs;
#endif
{
	char	*name,
		*rest;
	char	*type_c;
	int	type;
	char	*ArrayIndex;
	char	*EndIndex;

	type_c = next_arg(args, &rest);
	if (!type_c || !*type_c)
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}
	else
	{
		if (!my_strnicmp(type_c, "ASSIGN", 6))
			type = 1;
		else if (!my_strnicmp(type_c, "ALIAS", 5))
			type = 0;
		else
		{
			yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
			return;
		}
	}

	name = next_arg(rest, &rest);
	if (!name || !*name)
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}
	else
	{
		while ((ArrayIndex = (char *) index(name, '[')) != NULL)
		{
			*ArrayIndex++ = '.';
			if ((EndIndex = MatchingBracket(ArrayIndex, LEFT_BRACKET, RIGHT_BRACKET)) != NULL)
			{
				*EndIndex++ = '\0';
				strcat(name, EndIndex);
			}
			else
				break;
		}
	}
	if (!rest || !*rest)
	{
		yell("Usage: /STUB (ALIAS|ASSIGN) alias-name filename");
		return;
	}

	add_stub(type, name, rest);
}

/* built in expando functions */
static	char	*alias_line 		_((void)) { return m_strdup(get_input()); }
static	char	*alias_buffer 		_((void)) { return m_strdup(cut_buffer); }
static	char	*alias_time 		_((void)) { return m_strdup(update_clock(GET_TIME)); }
static	char	*alias_dollar 		_((void)) { return m_strdup("$"); }
static	char	*alias_detected 	_((void)) { return m_strdup(last_notify_nick); }
static	char	*alias_nick 		_((void)) { return m_strdup(get_server_nickname(curr_scr_win->server)); }
static	char	*alias_away 		_((void)) { return m_strdup(server_list[curr_scr_win->server].away); }
static	char	*alias_sent_nick 	_((void)) { return m_strdup((sent_nick) ? sent_nick : empty_string); }
static	char	*alias_recv_nick 	_((void)) { return m_strdup((recv_nick) ? recv_nick : empty_string); }
static	char	*alias_msg_body 	_((void)) { return m_strdup((sent_body) ? sent_body : empty_string); }
static	char	*alias_joined_nick 	_((void)) { return m_strdup((joined_nick) ? joined_nick : empty_string); }
static	char	*alias_public_nick 	_((void)) { return m_strdup((public_nick) ? public_nick : empty_string); }
static  char    *alias_show_realname 	_((void)) { return m_strdup(realname); }
static	char	*alias_version_str 	_((void)) { return m_strdup(irc_version); }
static	char	*alias_invite 		_((void)) { return m_strdup((invite_channel) ? invite_channel : empty_string); }
static	char	*alias_oper 		_((void)) { return m_strdup(get_server_operator(from_server) ?  get_string_var(STATUS_OPER_VAR) : empty_string); }
static	char	*alias_version 		_((void)) { return m_strdup(internal_version); }
static  char    *alias_online 		_((void)) { return m_sprintf("%ld",(long)start_time); }
static  char    *alias_idle 		_((void)) { return m_sprintf("%ld",time(NULL)-idle_time); }
static  char    *alias_show_userhost 	_((void)) { return m_sprintf("%s@%s",username, hostname); }
static	char	*alias_currdir  	_((void)) { return getcwd(NULL, MAXPATHLEN); }
static	char	*alias_current_numeric	_((void)) { return m_sprintf("%03d", -current_numeric); }

static	char	*alias_channel 		_((void)) 
{ 
	char	*tmp; 
	return m_strdup((tmp = get_channel_by_refnum(0)) ? tmp : "0");
}

static	char	*alias_server 		_((void)) 
{
	return m_strdup((parsing_server_index == -1) ?
		         get_server_itsname(parsing_server_index) :
		         (get_window_server(0) != -1) ?
			        get_server_itsname(get_window_server(0)) : empty_string);
}

static	char	*alias_query_nick 	_((void))
{
	char	*tmp;
	return m_strdup((tmp = query_nick()) ? tmp : empty_string);
}

static	char	*alias_target 		_((void))
{
	char	*tmp;
	return m_strdup((tmp = get_target_by_refnum(0)) ? tmp : empty_string);
}

static	char	*alias_cmdchar 		_((void))
{
	char	*cmdchars, tmp[2];

	if ((cmdchars = get_string_var(CMDCHARS_VAR)) == (char *) 0)
		cmdchars = DEFAULT_CMDCHARS;
	tmp[0] = cmdchars[0];
	tmp[1] = 0;
	return m_strdup(tmp);
}

static	char	*alias_chanop 		_((void))
{
	char	*tmp;
	return m_strdup(((tmp = get_channel_by_refnum(0)) && get_channel_oper(tmp, from_server)) ?
		"@" : empty_string);
}

static	char	*alias_modes 		_((void))
{
	char	*tmp;
	return m_strdup((tmp = get_channel_by_refnum(0)) ?
		get_channel_mode(tmp, from_server) : empty_string);
}

static	char	*alias_server_version  _((void))
{
	char	*s;
	return m_strdup((s = server_list[curr_scr_win->server].version_string) ?  s : empty_string);
}


/*	*	*	*	*	*	*	*	*	*
		These are the built-in functions.

	About 80 of them are here, the rest are in array.c.  All of the
	stock client's functions are supported, as well as about 60 more.
	Most of the 30 stock client's functions have been re-written for
	optimization reasons, and also to further distance ircii's code
	from EPIC.
 *	*	*	*	*	*	*	*	*	*/

/* 
 * These are defined to make the construction of the built-in functions
 * easier and less prone to bugs and unexpected behaviors.  As long as
 * you consistently use these macros to do the dirty work for you, you
 * will never have to do bounds checking as the macros do that for you. >;-) 
 *
 * Yes, i realize it makes the code slightly less efficient, but i feel that 
 * the cost is minimal compared to how much time i have spent over the last 
 * year debugging these functions and the fact i wont have to again. ;-)
 */
#define EMPTY empty_string
#define RETURN_EMPTY return m_strdup(EMPTY)
#define RETURN_IF_EMPTY(x) if (!x || !*x) RETURN_EMPTY
#define GET_INT_ARG(x, y) {RETURN_IF_EMPTY(y); x = atol(new_next_arg(y, &y));}
#define GET_FLOAT_ARG(x, y) {RETURN_IF_EMPTY(y); x = atof(new_next_arg(y, &y));}
#define GET_STR_ARG(x, y) {RETURN_IF_EMPTY(y); x = new_next_arg(y, &y);}
#define RETURN_STR(x) return m_strdup(x ? x : EMPTY)
#define RETURN_INT(x) return m_sprintf("%ld", x);

/*
 * Usage: $left(number text)
 * Returns: the <number> leftmost characters in <text>.
 * Example: $left(5 the quick brown frog) returns "the q"
 *
 * Note: the difference between $[10]foo and $left(10 foo) is that the former
 * is padded and the latter is not.
 */
#ifdef __STDC__
static char	*function_left(char *word)
#else
static char	*function_left(word)
	char	*word;
#endif
{
	long	count;

	GET_INT_ARG(count, word);
	RETURN_IF_EMPTY(word);

	if (strlen(word) > count)
		word[count] = 0;

	RETURN_STR(word);
}

/*
 * Usage: $right(number text)
 * Returns: the <number> rightmost characters in <text>.
 * Example: $right(5 the quick brown frog) returns " frog"
 */
#ifdef __STDC__
static char	*function_right(char *word)
#else
static char	*function_right(word)
	char	*word;
#endif
{
	long	count;

	GET_INT_ARG(count, word);
	RETURN_IF_EMPTY(word);

	if (strlen(word) > count)
		word += strlen(word) - count;

	RETURN_STR(word);
}

/*
 * Usage: $mid(start number text)
 * Returns: the <start>th through <start>+<number>th characters in <text>.
 * Example: $mid(3 4 the quick brown frog) returns " qui"
 *
 * Note: the first character is numbered zero.
 */
#ifdef __STDC__
static char	*function_mid(char *word)
#else
static char	*function_mid(word)
	char	*word;
#endif
{
	long	start, length;

	GET_INT_ARG(start, word);
	GET_INT_ARG(length, word);
	RETURN_IF_EMPTY(word);

	if (start < strlen(word))
	{
		word += start;
		if (length < strlen(word))
			word[length] = 0;
	}
	else
		word = EMPTY;

	RETURN_STR(word);
}

#ifdef HAVE_GETTIMEOFDAY
static unsigned long randl _((void))
{
	struct timeval tp1;
	gettimeofday(&tp1, NULL);
	return (unsigned long) tp1.tv_usec;
}
#endif

/* patch from Sarayan to make $rand() better */

#define RAND_A 16807L
#define RAND_M 2147483647L
#define RAND_Q 127773L
#define RAND_R 2836L

#ifdef __STDC__
static	unsigned long	randm(unsigned long l)
#else
static	unsigned long	randm(l)
	unsigned long	l;
#endif
{
/* If we have gettimeofday(2), we use it, because its 'leet.
	-- basically how it works is we take two samplings of 
	   gettimeofday().  Since who knows how long apart the samples
	   will be, we can be assured theyre *NOT* psudeo-random.  Then
	   we take the lower 16 bits of each sample, and then pack them
	   into a long, and viola, a random number.
	-- oh.  since we know that the usec field of timeval is microseconds,
	   its *always* more then 65535, so the lower 16 bits change at a
	   constant rate, which we cant say for that 17th bit. >;-)
 */
#ifdef HAVE_GETTIMEOFDAY
	unsigned long t1, t2, t;

	t1 = randl();
	t2 = randl();
	t = (t1 & 65535) * 65536 + (t2 & 65535);
	return t;
#else
	static	u_long	z = 0;
	long	t;

	if (!z)
		z = (u_long) getuid();
	if (!l)
	{
		t = RAND_A * (z % RAND_Q) - RAND_R * (z / RAND_Q);
		if (t > 0)
			z = t;
		else
			z = t + RAND_M;
		return (z >> 8) | ((z & 255) << 23);
	}
	else
	{
		if (l < 0)
			z = (u_long) getuid();
		else
			z = l;
		return 0;
	}
#endif
}

/*
 * Usage: $rand(max)
 * Returns: A random number from zero to max-1.
 * Example: $rand(10) might return any number from 0 to 9.
 */
#ifdef __STDC__
static char	*function_rand(char *word)
#else
static char	*function_rand(word)
	char	*word;
#endif
{
	long	tempin;

	GET_INT_ARG(tempin, word);
	if (tempin == 0)
		tempin = (unsigned long) -1;	/* This is cheating. :P */
	RETURN_INT(randm(0L) % tempin);
}

/*
 * Usage: $srand(seed)
 * Returns: Nothing.
 * Side effect: seeds the random number generater.
 * Note: the argument is ignored.
 */
#ifdef __STDC__
static 	char	*function_srand(char *word)
#else
static	char	*function_srand(word)
	char	*word;
#endif
{
	randm((long) time(NULL));
	RETURN_EMPTY;
}

/*
 * Usage: $time()
 * Returns: The number of seconds that has elapsed since Jan 1, 1970, GMT.
 * Example: $time() returned something around 802835348 at the time I
 * 	    wrote this comment.
 */
#ifdef __STDC__
static char	*function_time(char *input)
#else
static char	*function_time(input)
	char	*input;
#endif
{
	RETURN_INT(time(NULL));
}

/*
 * Usage: $stime(time)
 * Returns: The human-readable form of the date based on the <time> argument.
 * Example: $stime(1000) returns what time it was 1000 seconds from the epoch.
 * 
 * Note: $stime() is really useful when you give it the argument $time(), ala
 *       $stime($time()) is the human readable form for now.
 */
#ifdef __STDC__
static char	*function_stime (char *input)
#else
static char	* function_stime(input)
	char	*input;
#endif
{
	char	*result;
	time_t	ltime;

	GET_INT_ARG(ltime, input);
	result = m_strdup(ctime(&ltime));
	chop(result);
	RETURN_STR(result);
}

/*
 * Usage: $tdiff(seconds)
 * Returns: The time that has elapsed represented in days/hours/minutes/seconds
 *          corresponding to the number of seconds passed as the argument.
 * Example: $tdiff(3663) returns "1 hour 1 minute 3 seconds"
 */
#ifdef __STDC__
static char	*function_tdiff (char *input)
#else
static char	* function_tdiff(input)
	char	*input;
#endif
{
	time_t	ltime;
	time_t	days,
		hours,
		minutes,
		seconds;
	char	tmp[80];
	char	*tstr;

	GET_INT_ARG(ltime, input);

	seconds = ltime % 60;
	ltime = (ltime - seconds) / 60;
	minutes = ltime % 60;
	ltime = (ltime - minutes) / 60;
	hours = ltime % 24;
	days = (ltime - hours) / 24;
	tstr = tmp;
	if (days)
	{
		sprintf(tstr, "%ld day%s ", days, plural(days));
		tstr += strlen(tstr);
	}
	if (hours)
	{
		sprintf(tstr, "%ld hour%s ", hours, plural(hours));
		tstr += strlen(tstr);
	}
	if (minutes)
	{
		sprintf(tstr, "%ld minute%s ", minutes, plural(minutes));
		tstr += strlen(tstr);
	}
	if (seconds || (!days && !hours && !minutes))
	{
		sprintf(tstr, "%ld second%s", seconds, plural(seconds));
		tstr += strlen(tstr);
	}

	RETURN_STR(tmp);
}

/*
 * Usage: $index(characters text)
 * Returns: The number of leading characters in <text> that do not occur 
 *          anywhere in the <characters> argument.
 * Example: $index(f three fine frogs) returns 6 (the 'f' in 'fine')
 *          $index(frg three fine frogs) returns 2 (the 'r' in 'three')
 */
#ifdef __STDC__
static char	*function_index (char *input)
#else
static char	* function_index(input)
	char	*input;
#endif
{
	char	*schars;
	char	*iloc;

	GET_STR_ARG(schars, input);
	iloc = sindex(input, schars);
	RETURN_INT(iloc ? iloc - input : -1);
}

/*
 * Usage: $rindex(characters text)
 * Returns: The number of leading characters in <text> that occur before the
 *          *last* occurance of any of the characters in the <characters> 
 *          argument.
 * Example: $rindex(f three fine frogs) returns 12 (the 'f' in 'frogs')
 *          $rindex(frg three fine frogs) returns 15 (the 'g' in 'frogs')
 */
#ifdef __STDC__
static char	*function_rindex (char *word)
#else
static char	* function_rindex(word)
	char	*word;
#endif
{
	char	*chars, *last;

	/* need to find out why ^x doesnt work */
	GET_STR_ARG(chars, word);
	last = rsindex(word + strlen(word) - 1, word, chars);
	RETURN_INT(last ? last - word : -1);
}

/*
 * Usage: $match(pattern list of words)
 * Returns: if no words in the list match the pattern, it returns 0.
 *	    Otherwise, it returns the number of the word that most
 *	    exactly matches the pattern (first word is numbered one)
 * Example: $match(f*bar foofum barfoo foobar) returns 3
 *	    $match(g*ant foofum barfoo foobar) returns 0
 *
 * Note: it is possible to embed spaces inside of a word or pattern simply
 *       by including the entire word or pattern in quotation marks. (")
 */
#ifdef __STDC__
static char	*function_match (char *input)
#else
static char	* function_match(input)
	char	*input;
#endif
{
	char	*pattern, 	*word;
	long	current_match,	best_match = 0,	match = 0, match_index = 0;

	GET_STR_ARG(pattern, input);

	while (input && *input)
	{
		match_index++;
		GET_STR_ARG(word, input);
		if ((current_match = wild_match(pattern, word)) > best_match)
		{
			match = match_index;
			best_match = current_match;
		}
	}

	RETURN_INT(match);
}

/*
 * Usage: $rmatch(word list of patterns)
 * Returns: if no pattern in the list matches the word, it returns 0.
 *	    Otherwise, it returns the number of the pattern that most
 *	    exactly matches the word (first word is numbered one)
 * Example: $rmatch(foobar f*bar foo*ar g*ant) returns 2 
 *	    $rmatch(booya f*bar foo*ar g*ant) returns 0
 * 
 * Note: It is possible to embed spaces into a word or pattern simply by
 *       including the entire word or pattern within quotation marks (")
 */
#ifdef __STDC__
static char	*function_rmatch (char *input)
#else
static char	*function_rmatch (input)
	char	*input;
#endif
{
	char	*pattern,	*word;
	int	current_match,	best_match = 0,	match = 0, rmatch_index = 0;

	GET_STR_ARG(word, input);

	while (input && *input)
	{
		rmatch_index++;
		GET_STR_ARG(pattern, input);
		if ((current_match = wild_match(pattern, word)) > best_match)
		{
			match = rmatch_index;
			best_match = current_match;
		}
	}

	RETURN_INT(match);
}

/*
 * Usage: $userhost()
 * Returns: the userhost (if any) of the most previously recieved message.
 * Caveat: $userhost() changes with every single line that appears on
 *         your screen, so if you want to save it, you will need to assign
 *         it to a variable.
 */
#ifdef __STDC__
static char	*function_userhost (char *input)
#else
static char	*function_userhost (input)
	char	*input;
#endif
{
	RETURN_STR(FromUserHost);
}

/* 
 * Usage: $strip(characters text)
 * Returns: <text> with all instances of any characters in the <characters>
 *	    argument removed.
 * Example: $strip(f free fine frogs) returns "ree ine rogs"
 *
 * Note: it can be difficult (actually, not possible) to remove spaces from
 *       a string using this function.  To remove spaces, simply use this:
 *		$tr(/ //$text)
 *
 *	 Actually, i recommend not using $strip() at all and just using
 *		$tr(/characters//$text)
 *	 (but then again, im biased. >;-)
 */
#ifdef __STDC__
static char	*function_strip (char *input)
#else
static char	* function_strip(input)
	char	*input;
#endif
{
	char	*result;
	char	*chars;
	char	*cp, *dp;

	GET_STR_ARG(chars, input);
	RETURN_IF_EMPTY(input);

	result = (char *)new_malloc(strlen(input) + 1);
	for (cp = input, dp = result; *cp; cp++)
	{
		/* This is expensive -- gotta be a better way */
		if (!index(chars, *cp))
			*dp++ = *cp;
	}
	*dp = '\0';

	return result;		/* DONT USE RETURN_STR HERE! */
}

/*
 * Usage: $encode(text)
 * Returns: a string, uniquely identified with <text> such that the string
 *          can be used as a variable name.
 * Example: $encode(fe fi fo fum) returns "GGGFCAGGGJCAGGGPCAGGHFGN"
 *
 * Note: $encode($decode(text)) returns text (most of the time)
 *       $decode($encode(text)) also returns text.
 */
#ifdef __STDC__
static char	*function_encode (unsigned char *input)
#else
static char	* function_encode(input)
	unsigned char	*input;
#endif
{
	char	*result;
	int	i = 0;

	result = (char *)new_malloc(strlen(input) * 2 + 1);
	while (*input)
	{
		result[i++] = (*input >> 4) + 0x41;
		result[i++] = (*input & 0x0f) + 0x41;
		input++;
	}
	result[i] = '\0';

	return result;		/* DONT USE RETURN_STR HERE! */
}


/*
 * Usage: $decode(text)
 * Returns: If <text> was generated with $encode(), it returns the string
 *          you originally encoded.  If it wasnt, you will probably get
 *	    nothing useful in particular.
 * Example: $decode(GGGFCAGGGJCAGGGPCAGGHFGN) returns "fe fi fo fum"
 *
 * Note: $encode($decode(text)) returns "text"
 *       $decode($encode(text)) returns "text" too.
 *
 * Note: Yes.  $decode(plain-text) does compress the data by a factor of 2.
 *       But it ignores non-ascii text, so use this as compression at your
 *	 own risk and peril.
 */
#ifdef __STDC__
static char	*function_decode (unsigned char *input)
#else
static char	* function_decode(input)
	unsigned char	*input;
#endif
{
	char	*result;
	int	i = 0;

	result = (char *)new_malloc(strlen(input) / 2 + 1);

	while (input[0] && input[1])
	{
		/* oops, this isnt quite right. */
		result[i] = ((input[0] - 0x41) << 4) | (input[1] - 0x41);
		input += 2;
		i++;
	}
	result[i] = '\0';

	return result;		/* DONT USE RETURN_STR HERE! */
}


/*
 * Usage: $ischannel(text)
 * Returns: If <text> could be a valid channel name, 1 is returned.
 *          If <text> is an illegal channel name, 0 is returned.
 *
 * Note: Contrary to popular belief, this function does NOT determine
 * whether a given channel name is in use!
 */
#ifdef __STDC__
static char	*function_ischannel (char *input)
#else
static char	* function_ischannel(input)
	char	*input;
#endif
{
	RETURN_INT(is_channel(input));
}

/*
 * Usage: $ischanop(nick channel)
 * Returns: 1 if <nick> is a channel operator on <channel>
 *          0 if <nick> is not a channel operator on <channel>
 *			* O R *
 *	      if you are not on <channel>
 *
 * Note: Contrary to popular belief, this function can only tell you
 *       who the channel operators are for channels you are already on!
 *
 * Boo Hiss:  This should be $ischanop(channel nick <nick...nick>)
 *	      and return a list (1 1 ... 0), which would allow us to
 *	      call is_chanop() without ripping off the nick, and allow 
 *	      us to abstract is_chanop() to take a list. oh well... 
 *	      Too late to change it now. :/
 */
#ifdef __STDC__
static char	*function_ischanop (char *input)
#else
static char	*function_ischanop(input)
	char	*input;
#endif
{
	char	*nick;

	GET_STR_ARG(nick, input);
	RETURN_INT(is_chanop(input, nick));
}


/*
 * Usage: $word(number text)
 * Returns: the <number>th word in <text>.  The first word is numbered zero.
 * Example: $word(3 one two three four five) returns "four" (think about it)
 */
#ifdef __STDC__
static char	*function_word (char *word)
#else
static char	* function_word(word)
	char	*word;
#endif
{
	int	cvalue;
	char	*w_word;

	GET_INT_ARG(cvalue, word);
	if (cvalue < 0)
		RETURN_EMPTY;

	while (cvalue-- > 0)
		GET_STR_ARG(w_word, word);

	GET_STR_ARG(w_word, word);
	RETURN_STR(w_word);
}


/*
 * Usage: $winnum()
 * Returns: the index number for the current window
 * 
 * Note: returns -1 if there are no windows open (ie, in dumb mode)
 */
#ifdef __STDC__
static char	*function_winnum (char *input)
#else
static char	* function_winnum(input)
	char	*input;
#endif
{
	RETURN_INT(curr_scr_win ? curr_scr_win->refnum : -1);
}

static
#ifdef __STDC__
char	*function_winnam (char *input)
#else
char	* function_winnam(input)
	char	*input;
#endif
{
	if (curr_scr_win)
		RETURN_STR(curr_scr_win->name);
	else
		RETURN_EMPTY;
}

static
#ifdef __STDC__
char	*function_connect (char *input)
#else
char	* function_connect(input)
	char	*input;
#endif
{
	char	*host;
	int	port;

	GET_STR_ARG(host, input);
	GET_INT_ARG(port, input);

	return dcc_raw_connect(host, port);	/* DONT USE RETURN_STR HERE! */
}


static
#ifdef __STDC__
char	*function_listen (char *input)
#else
char	* function_listen(input)
	char	*input;
#endif
{
	int	port;
	char	*result;

	GET_INT_ARG(port, input);
	result = dcc_raw_listen(port);
	RETURN_STR(result);			/* DONT REMOVE RESULT! */
}

static
#ifdef __STDC__
char	*function_toupper (char *input)
#else
char	* function_toupper(input)
	char	*input;
#endif
{
	char	*new = (char *) 0,
		*ptr;

	RETURN_IF_EMPTY(input);

	malloc_strcpy(&new, input);
	for (ptr = new; *ptr; ptr++)
		*ptr = islower(*ptr) ? toupper(*ptr) : *ptr;
	return new;
}

static
#ifdef __STDC__
char	*function_tolower (char *input)
#else
char	* function_tolower(input)
	char	*input;
#endif
{
	char	*new = (char *) 0,
		*ptr;

	RETURN_IF_EMPTY(input);

	malloc_strcpy(&new, input);
	for (ptr = new; *ptr; ptr++)
		*ptr = isupper(*ptr) ? tolower(*ptr) : *ptr;
	return new;
}

static
#ifdef __STDC__
char	*function_curpos (char *input)
#else
char	* function_curpos(input)
	char	*input;
#endif
{
	RETURN_INT(current_screen->buffer_pos);
}

static
#ifdef __STDC__
char	*function_channels (char *input)
#else
char	* function_channels(input)
	char	*input;
#endif
{
	long	winnum;
	Window  *window = curr_scr_win;

	if (input)
	{
		GET_INT_ARG(winnum, input);
		window = get_window_by_refnum(winnum);
	}

	return create_channel_list(window);	/* DONT USE RETURN_STR HERE! */
}

static
#ifdef __STDC__
char	*function_servers (char *input)
#else
char	* function_servers(input)
	char	*input;
#endif
{
	return create_server_list();		/* DONT USE RETURN_STR HERE! */
}

static
#ifdef __STDC__
char	*function_pid(char *input)
#else
char	* function_pid(input)
	char	*input;
#endif
{
	RETURN_INT(getpid());
}

static
#ifdef __STDC__
char	*function_ppid (char *input)
#else
char	* function_ppid(input)
	char	*input;
#endif
{
	RETURN_INT(getppid());
}



/* The new "added" functions */

/* $before(chars string of text)
 * returns the part of "string of text" that occurs before the
 * first instance of any character in "chars"
 * EX:  $before(! hop!jnelson@iastate.edu) returns "hop"
 */
static
#ifdef __STDC__
char	*function_before (char *word)
#else
char    *function_before(word)
char    *word;
#endif
{
	char	*pointer = (char *) 0;
	char	*chars;
	char	*tmp;
	long	numint;

	GET_STR_ARG(tmp, word);			/* DONT DELETE TMP! */
	numint = atol(tmp);

	if (numint)
	{
		GET_STR_ARG(chars, word);
	}
	else
	{
		numint = 1;
		chars = tmp;
	}

	if (numint < 0)
		pointer = word + strlen(word) - 1;

	pointer = search(word, &pointer, chars, numint);

	if (!pointer)
		RETURN_EMPTY;

	*pointer = '\0';
	RETURN_STR(word);
}

/* $after(chars string of text)
 * returns the part of "string of text" that occurs after the 
 * first instance of any character in "chars"
 * EX: $after(! hop!jnelson@iastate.edu)  returns "jnelson@iastate.edu"
 */
static
#ifdef __STDC__
char	*function_after (char *word)
#else
char 	*function_after(word)
char	*word;
#endif
{
	char	*chars;
	char	*pointer = (char *) 0;
	char 	*tmp;
	long	numint;

	GET_STR_ARG(tmp, word);
	numint = atol(tmp);

	if (numint)
		chars = new_next_arg(word, &word);
	else
	{
		numint = 1;
		chars = tmp;
	}

	if (numint < 0)
		pointer = word + strlen(word) - 1;

	pointer = search(word, &pointer, chars, numint);

	if (!pointer)
		RETURN_EMPTY;

	RETURN_STR(pointer + 1);
}

/* $leftw(num string of text)
 * returns the left "num" words in "string of text"
 * EX: $leftw(3 now is the time for) returns "now is the"
 */
static
#ifdef __STDC__
char	*function_leftw (char *word)
#else
char    *function_leftw(word)
char    *word;
#endif
{
	int value;
 
	GET_INT_ARG(value, word);
	if (value < 1)
		RETURN_EMPTY;

	return (extract(word, 0, value-1));	/* DONT USE RETURN_STR HERE! */
}

/* $rightw(num string of text)
 * returns the right num words in "string of text"
 * EX: $rightw(3 now is the time for) returns "the time for"
 */
static
#ifdef __STDC__
char	*function_rightw (char *word)
#else
char    *function_rightw(word)
char    *word;
#endif
{
	int     value;

	GET_INT_ARG(value, word);
	if (value < 1)
		RETURN_EMPTY;
		
	return extract2(word, -value, EOS); 
}


/* $midw(start num string of text)
 * returns "num" words starting at word "start" in the string "string of text"
 * NOTE: The first word is word #0.
 * EX: $midw(2 2 now is the time for) returns "the time"
 */
static
#ifdef __STDC__
char	*function_midw (char *word)
#else
char    *function_midw(word)
char    *word;
#endif
{
	int     start, num;

	GET_INT_ARG(start, word);
	GET_INT_ARG(num, word);

	if (num < 1)
		RETURN_EMPTY;

	return extract(word, start, (start + num - 1));
}

/* $notw(num string of text)
 * returns "string of text" with word number "num" removed.
 * NOTE: The first word is numbered 0.
 * EX: $notw(3 now is the time for) returns "now is the for"
 */
static
#ifdef __STDC__
char	*function_notw (char *word)
#else
char    *function_notw(word)
char    *word;
#endif
{
	char    *booya = (char *)0;
	int     where;
	
	GET_INT_ARG(where, word);

	/* An illegal word simply returns the string as-is */
	if (where < 0)
		RETURN_STR(word);

	if (where > 0)
	{
		char *part1, *part2;
		part1 = extract(word, 0, (where - 1));
		part2 = extract(word, (where + 1), EOS);
		booya = m_strdup(part1);
		if (part2 && *part2)
		{
			malloc_strcat(&booya, " ");
			malloc_strcat(&booya, part2);
		}
		new_free(&part1);
		new_free(&part2);
	}
	else /* where == 0 */
		booya = extract(word, 1, EOS);

	return booya;				/* DONT USE RETURN_STR HERE! */
}

/* $restw(num string of text)
 * returns "string of text" that occurs starting with and including
 * word number "num"
 * NOTE: the first word is numbered 0.
 * EX: $restw(3 now is the time for) returns "time for"
 */
static
#ifdef __STDC__
char	*function_restw(char *word)
#else
char    *function_restw(word)
char    *word;
#endif
{
	int     where;
	
	GET_INT_ARG(where, word);
	if (where < 0)
		RETURN_EMPTY;
	return extract(word, where, EOS);
}

/* $remw(word string of text)
 * returns "string of text" with the word "word" removed
 * EX: $remw(the now is the time for) returns "now is time for"
 */
static
#ifdef __STDC__
char	*function_remw (char *word)
#else
char    *function_remw(word)
char    *word;
#endif
{
	char    *booya = (char *) 0;
	char    *blah = (char *) 0;
	char    *lame = (char *) 0, *plame;
	int	where;

	lame = m_strdup(word);
	plame = lame;
	blah = function_rmatch(word);
	where = atoi(blah);

	/* No matches, return the string as-is */
	if (where == 0)
	{
		new_next_arg(lame, &lame);
		booya = m_strdup(lame);
	}
	else
	{
		char *str1, *str2;
		str1 = extract(lame, 1, (where - 1));
		str2 = extract(lame, (where + 1), EOS);
		booya = m_strdup(str1);
		if (str2 && *str2)
		{
			malloc_strcat(&booya, " ");
			malloc_strcat(&booya, str2);
		}
		new_free(&str1);
		new_free(&str2);
	}
	new_free(&blah);
	new_free(&plame);
	return booya;				/* DONT USE RETURN_STR HERE! */
}

/* $insertw(num word string of text)
 * returns "string of text" such that "word" is the "num"th word
 * in the string.
 * NOTE: the first word is numbered 0.
 * EX: $insertw(3 foo now is the time for) returns "now is the foo time for"
 */
static
#ifdef __STDC__
char	*function_insertw (char *word)
#else
char    *function_insertw(word)
char    *word;
#endif
{
	int     where;
	char    *what;
	char    *booya=(char *)0;
	char 	*str1, *str2;

	GET_INT_ARG(where, word);
	
	/* If the word goes at the front of the string, then it
	   already is: return it. ;-) */
	if (where < 1)
		booya = m_strdup(word);
	else
	{
		GET_STR_ARG(what, word);
		str1 = extract(word, 0, (where - 1));
		str2 = extract(word, where, EOS);
		booya = m_strdup(str1);
		if (str1 && *str1)
			malloc_strcat(&booya, " ");
		malloc_strcat(&booya, what);
		if (str2 && *str2)
		{
			malloc_strcat(&booya, " ");
			malloc_strcat(&booya, str2);
		}
		new_free(&str1);
		new_free(&str2);
	}

	return booya;				/* DONT USE RETURN_STR HERE! */
}

/* $chngw(num word string of text)
 * returns "string of text" such that the "num"th word is removed 
 * and replaced by "word"
 * NOTE: the first word is numbered 0
 * EX: $chngw(3 foo now is the time for) returns "now is the foo for"
 */
static
#ifdef __STDC__
char	*function_chngw (char *word)
#else
char    *function_chngw(word)
char    *word;
#endif
{
	int     which;
	char    *what;
	char    *booya=(char *)0;
	char	*str1, *str2;
	
	GET_INT_ARG(which, word);
	GET_STR_ARG(what, word);

	if (which < 0)
		RETURN_STR(word);

	/* hmmm. if which is 0, extract does the wrong thing. */
	str1 = extract(word, 0, which - 1);
	str2 = extract(word, which + 1, EOS);
	booya = m_strdup(str1);
	if (str1 && *str1)
		malloc_strcat(&booya, " ");
	malloc_strcat(&booya, what);
	if (str2 && *str2)
	{
		malloc_strcat(&booya, " ");
		malloc_strcat(&booya, str2);
	}
	new_free(&str1);
	new_free(&str2);

	return (booya);
}


/* $common (string of text / string of text)
 * Given two sets of words seperated by a forward-slash '/', returns
 * all words that are found in both sets.
 * EX: $common(one two three / buckle my two shoe one) returns "one two"
 * NOTE: returned in order found in first string.
 */
static
#ifdef __STDC__
char	*function_common (char *word)
#else
char    *function_common(word)
char    *word;
#endif
{
	char    *left = (char *) 0,
		*right = (char *) 0,
		*blah = (char *) 0,
		*booya = (char *) 0;

	left = word;
	if ((right = index(word,'/')) == NULL)
		RETURN_EMPTY;

	do
		*(right++) = '\0';
	while (*right && isspace(*right));

	while ((blah = new_next_arg(left, &left)))
	{
		if (stristr(right, blah))
		{
			if (booya)
				malloc_strcat(&booya, " ");
			malloc_strcat(&booya, blah);
		}
	}
	if (!booya)
		RETURN_EMPTY;

	return (booya);				/* DONT USE RETURN_STR HERE! */
}

/* $diff(string of text / string of text)
 * given two sets of words, seperated by a forward-slash '/', returns
 * all words that are not found in both sets
 * EX: $diff(one two three / buckle my two shoe)
 * returns "one two three buckle my shoe"
 */
static
#ifdef __STDC__
char	*function_diff (char *word)
#else
char    *function_diff(word)
char    *word;
#endif
{
	char    *left = (char *) 0;
	char    *right = (char *) 0;
	char    *booya = m_strdup(empty_string);
	char    *blah = (char *) 0;
	char    *leftcopy = (char *) 0;

	left = word;
	if ((right = index(word,'/')) == (char *) 0)
		RETURN_EMPTY;

	do 
		*(right++) = '\0';
	while (isspace(*right));

	leftcopy = m_strdup(left);

	while ((blah = new_next_arg(left, &left)))
	{
		if (!stristr(right, blah)) 
		{
			if (booya)
			{
				if (!stristr(booya, blah))
				{
					if (booya && *booya)
						malloc_strcat(&booya, " ");
					malloc_strcat(&booya, blah);
				}
			}
			else
				malloc_strcpy(&booya, blah);
		}
	}
	/* ok.. right now we have everything in the left side that
	 * isnt in the right side.  We now need to extract the things
	 * in the right side that arent in the left side... hmmmm
	 */
	while ((blah = new_next_arg(right, &right)))
	{
		/* if the word ISNT already on the list */
		if (!stristr(booya, blah))
		{
			/* if the word ISNT on the left side */
			if (!stristr(leftcopy, blah))
			{
				/* put it in */
				if (booya && *booya)
					malloc_strcat(&booya, " ");
				malloc_strcat(&booya, blah);
			}
		}
	}
	new_free(&leftcopy);
	return (booya);
}

/* $pattern(pattern string of words)
 * given a pattern and a string of words, returns all words that
 * are matched by the pattern
 * EX: $pattern(f* one two three four five) returns "four five"
 */
static
#ifdef __STDC__
char	*function_pattern (char *word)
#else
char    *function_pattern(word)
char    *word;
#endif
{
	char    *blah;
	char    *booya = m_strdup(empty_string);
	char    *pattern;

	if ((pattern = new_next_arg(word, &word)))
	{
		while (((blah = new_next_arg(word, &word)) != NULL))
		{
			if (wild_match(pattern, blah))
			{
				if (*booya)
					malloc_strcat(&booya, " ");
				malloc_strcat(&booya, blah);
			}
		}
		return (booya);
	} 
	else 
		return m_strdup(empty_string);	
}

/* $filter(pattern string of words)
 * given a pattern and a string of words, returns all words that are 
 * NOT matched by the pattern
 * $filter(f* one two three four five) returns "one two three"
 */
static
#ifdef __STDC__
char	*function_filter (char *word)
#else
char    *function_filter(word)
char    *word;
#endif
{
	char    *blah;
	char    *booya = m_strdup(empty_string);
	char    *pattern;

	if ((pattern = new_next_arg(word, &word)))
	{
		while ((blah = new_next_arg(word, &word)) != NULL)
		{
			if (!wild_match(pattern, blah))
			{
				if (*booya)
					malloc_strcat(&booya, " ");
				malloc_strcat(&booya, blah);
			}
		}
		return (booya);
	} 
	else 
		return m_strdup(empty_string);
}

/* $rpattern(word list of patterns)
 * Given a word and a list of patterns, return all patterns that
 * match the word.
 * EX: $rpattern(jnelson@iastate.edu *@* jnelson@* f*@*.edu)
 * returns "*@* jnelson@*"
 */
static
#ifdef __STDC__
char	*function_rpattern (char *word)
#else
char    *function_rpattern(word)
char    *word;
#endif
{
	char    *blah;
	char    *booya = m_strdup(empty_string);
	char    *pattern;

	if ((blah = new_next_arg(word, &word)))
	{
		while ((pattern = new_next_arg(word, &word)) != NULL)
		{
			if (wild_match(pattern, blah))
			{
				if (*booya)
					malloc_strcat(&booya, " ");
				malloc_strcat(&booya, pattern);
			}
		}
		return (booya);
	} 
	else 
		return m_strdup(empty_string);
}

/* $rfilter(word list of patterns)
 * given a word and a list of patterns, return all patterns that
 * do NOT match the word
 * EX: $rfilter(jnelson@iastate.edu *@* jnelson@* f*@*.edu)
 * returns "f*@*.edu"
 */
static
#ifdef __STDC__
char	*function_rfilter (char *word)
#else
char    *function_rfilter(word)
char    *word;
#endif
{
	char    *blah;
	char    *booya = m_strdup(empty_string);
	char    *pattern;

	if ((blah = new_next_arg(word, &word)))
	{
		while ((pattern = new_next_arg(word, &word)) != NULL)
		{
			if (!wild_match(pattern, blah))
			{
				if (*booya)
					malloc_strcat(&booya, " ");
				malloc_strcat(&booya, pattern);
			 }
		}
		return (booya);
	} 
	else 
		return m_strdup(empty_string);
}

/* $copattern(pattern var_1 var_2)
 * Given a pattern and two variable names, it returns all words
 * in the variable_2 corresponding to any words in variable_1 that
 * are matched by the pattern
 * EX: @nicks = [hop IRSMan skip]
 *     @userh = [jnelson@iastate.edu irsman@iastate.edu sanders@rush.cc.edu]
 *     $copattern(*@iastate.edu userh nicks) 
 *	returns "hop IRSMan"
 */
static
#ifdef __STDC__
char	*function_copattern (char *word)
#else
char    *function_copattern(word)
char    *word;
#endif
{
	char	*booya = (char *) 0,
		*pattern = (char *) 0,
		*firstl = (char *) 0, *firstlist = (char *) 0, *firstel = (char *) 0,
		*secondl = (char *) 0, *secondlist = (char *) 0, *secondel = (char *) 0;
	char 	*sfirstl, *ssecondl;

	GET_STR_ARG(pattern, word);
	GET_STR_ARG(firstlist, word);
	GET_STR_ARG(secondlist, word);

	firstl = find_inline(firstlist);
	secondl = find_inline(secondlist);
	sfirstl = firstl;
	ssecondl = secondl;

	while ((firstel = new_next_arg(firstl, &firstl)))
	{
		if (!(secondel = new_next_arg(secondl, &secondl)))
		{
			(void) (booya || (booya = m_strdup(empty_string)));
			break;
		}

		if (wild_match(pattern, firstel))
		{
			if (booya)
				malloc_strcat(&booya, " ");
			malloc_strcat(&booya, secondel);
		}
	}
	new_free(&sfirstl);
	new_free(&ssecondl);

	if (!booya) 
		RETURN_EMPTY;

	return (booya);
}

/* $beforew(pattern string of words)
 * returns the portion of "string of words" that occurs before the 
 * first word that is matched by "pattern"
 * EX: $beforew(three one two three o leary) returns "one two"
 */
static
#ifdef __STDC__
char	*function_beforew (char *word)
#else
char    *function_beforew(word)
char    *word;
#endif
{
	int     where;
	char	*lame = (char *) 0;
	char	*placeholder;

	lame = m_strdup(word);
	placeholder = function_rmatch(word);
	where = atoi(placeholder);

	new_free(&placeholder);

	if (where < 1)
	{
		new_free(&lame);
		RETURN_EMPTY;
	}
	placeholder = extract(lame, 1, where - 1);
	new_free(&lame);
	return placeholder;
}
		
/* Same as above, but includes the word being matched */
static
#ifdef __STDC__
char	*function_tow (char *word)
#else
char    *function_tow(word)
char    *word;
#endif
{
	int     where;
	char	*lame = (char *) 0;
	char	*placeholder;

	lame = m_strdup(word);
	placeholder = function_rmatch(word);
	where = atoi(placeholder);

	new_free(&placeholder);

	if (where < 1)
	{
		new_free(&lame);
		RETURN_EMPTY;
	}
	placeholder = extract(lame, 1, where);
	new_free(&lame);
	return placeholder;
}

/* Returns the string after the word being matched */
static
#ifdef __STDC__
char	*function_afterw (char *word)
#else
char    *function_afterw(word)
char    *word;
#endif
{
	int     where;
	char	*lame = (char *) 0;
	char	*placeholder;

	lame = m_strdup(word);
	placeholder = function_rmatch(word);
	where = atoi(placeholder);

	new_free(&placeholder);

	if (where < 1)
	{
		new_free(&lame);
		RETURN_EMPTY;
	}
	placeholder = extract(lame, where + 1, EOS);
	new_free(&lame);
	return placeholder;
}

/* Returns the string starting with the word being matched */
static
#ifdef __STDC__
char	*function_fromw (char *word)
#else
char    *function_fromw(word)
char    *word;
#endif
{
	int     where;
	char 	*lame = (char *) 0;
	char	*placeholder;

	lame = m_strdup(word);
	placeholder = function_rmatch(word);
	where = atoi(placeholder);

	new_free(&placeholder);

	if (where < 1)
	{
		new_free(&lame);
		RETURN_EMPTY;
	}

	placeholder = extract(lame, where, EOS);
	new_free(&lame);
	return placeholder;
}

/* Cut and paste a string */
static
#ifdef __STDC__
char	*function_splice (char *word)
#else
char    *function_splice(word)
char    *word;
#endif
{
	char    *variable;
	int	start;
	int	length;
	char 	*left_part = NULL;
	char	*middle_part = NULL;
	char	*right_part = NULL;
	char	*old_value = NULL;
	char	*new_value = NULL;
	int 	old_display = window_display;
	int	num_words;

	GET_STR_ARG(variable, word);
	GET_INT_ARG(start, word);
	GET_INT_ARG(length, word);

	old_value = find_inline(variable);
	num_words = word_count(old_value);

	if (start < 0)
	{
		if ((length += start) <= 0)
			RETURN_EMPTY;
		start = 0;
	}

	if (start >= num_words)
	{
		left_part = m_strdup(old_value);
		middle_part = m_strdup(empty_string);
		right_part = m_strdup(empty_string);
	}

	else if (start + length >= num_words)
	{
		left_part = extract(old_value, 0, start - 1);
		middle_part = extract(old_value, start, EOS);
		right_part = m_strdup(empty_string);
	}

	else
	{
		left_part = extract(old_value, 0, start - 1);
		middle_part = extract(old_value, start, start + length - 1);
		right_part = extract(old_value, start + length, EOS);
	}

	new_value = NULL;
	malloc_strcpy(&new_value, left_part);
	if (new_value && *new_value && word && *word)
		malloc_strcat(&new_value, " ");
	if (word && *word)
		malloc_strcat(&new_value, word);
	if (new_value && *new_value && *right_part && *right_part)
		malloc_strcat(&new_value, " ");
	if (right_part && *right_part)
		malloc_strcat(&new_value, right_part);

	window_display = 0;
	add_alias(VAR_ALIAS, variable, new_value);
	window_display = old_display;

	new_free(&old_value);
	new_free(&new_value);
	new_free(&left_part);
	new_free(&right_part);
	return middle_part;
}
	
static
#ifdef __STDC__
char	*function_numonchannel (char * word)
#else
char    *function_numonchannel(word)
char    *word;
#endif
{
	char            *channel = (char *) 0;
	ChannelList	*chan = (ChannelList *) 0;
	NickList	*tmp = (NickList *) 0;
	int             counter = 0;
	
	channel = new_next_arg(word, &word);
	if ((channel == (char *) 0) || (!is_channel(channel)) ||
	    (!(chan = lookup_channel(channel, curr_scr_win->server,0))))
		RETURN_INT(0);

	for (tmp = chan->nicks; tmp; tmp = tmp->next)
		counter++;

	RETURN_INT(counter);
}
	
static
#ifdef __STDC__
char	*function_onchannel (char *word)
#else
char    *function_onchannel(word)
char    *word;
#endif
{
	char		channel[255];	/* max channel name size */
	ChannelList     *chan = (ChannelList *) 0;
	NickList        *tmp = (NickList *) 0;
	char            *nicks = (char *) 0;

	strmcpy(channel, new_next_arg(word, &word), 254);
	if ((*channel == 0) || (!strcmp(channel, "*")))
		strmcpy(channel, get_channel_by_refnum(0), 254);

	if (*channel == 0 || *channel == '0')
		RETURN_EMPTY;

	if (!is_channel(channel))   /* must be a new onchannel */
	{
		char nick[255];

		strmcpy(nick, channel, 254);
		strmcpy(channel, new_next_arg(word, &word), 254);
		if (*channel)
			RETURN_INT(is_on_channel(channel, nick) ? 1 : 0)
		else
			RETURN_INT(0);
	}

	chan = lookup_channel(channel,curr_scr_win->server,0);
	if (chan == (ChannelList *) 0)
		RETURN_EMPTY;

	for (tmp = chan->nicks; tmp; tmp = tmp->next)
	{
		if (nicks)
		    malloc_strcat(&nicks, " ");
		malloc_strcat(&nicks, tmp->nick);
	}
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
	return(nicks);
}

static
#ifdef __STDC__
char    *function_chops(char *word)
#else
char    *function_chops(word)
char    *word;
#endif
{
	char		channel[255];	/* maximum size of channel name */
	ChannelList     *chan;
	NickList        *tmp;
	char            *nicks = (char *) 0;

	strmcpy(channel, new_next_arg(word, &word), 254);
	if ((*channel == 0) || (!strcmp(channel, "*")))
		strmcpy(channel, get_channel_by_refnum(0), 254);

	if (*channel == 0 || *channel == '0')
		RETURN_EMPTY;

	chan = lookup_channel(channel,curr_scr_win->server,0);
	if (!chan)
		RETURN_EMPTY;

	for (tmp = chan->nicks; tmp; tmp = tmp->next)
	{
		if (tmp->chanop == 1) 
		{
			if (nicks)
			    malloc_strcat(&nicks, " ");
			malloc_strcat(&nicks, tmp->nick);
		}
	}
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
	return(nicks);
}        

static
#ifdef __STDC__
char    *function_nochops(char *word)
#else
char    *function_nochops(word)
char    *word;
#endif
{
	ChannelList     *chan;
	NickList        *tmp;
	char            *nicks = (char *)0;
	char            channel[255];

	strmcpy(channel, new_next_arg(word, &word), 254);
	if ((*channel == 0) || (!strcmp(channel, "*")))
		strmcpy(channel, get_channel_by_refnum(0), 254);

	if (*channel == 0 || *channel == '0')
		RETURN_EMPTY;

	chan = lookup_channel(channel,curr_scr_win->server,0);
	if (!chan)
		RETURN_EMPTY;

	for (tmp = chan->nicks; tmp; tmp = tmp->next)
	{
		if (tmp->chanop == 0) 
		{
			if (nicks)
			    malloc_strcat(&nicks, " ");
			malloc_strcat(&nicks, tmp->nick);
		}
	}
	if (!nicks)
		malloc_strcpy(&nicks, empty_string);
	return(nicks);
}

static
#ifdef __STDC__
char    *function_key(char *word)
#else
char    *function_key(word)
char    *word;
#endif
{
	char	channel[257];
	char    *booya = (char *) 0;

	strmcpy(channel, new_next_arg(word, &word), 256);
	do
	{
		char    *key;

		/* If we are at the end of the list */
		if (!*channel)
		{
			/* If there were other channels, stop */
			if (booya)
				break;

			/* Empty list -- use current channel */
			else
				strmcpy(channel, get_channel_by_refnum(0), 256);
		}

		/* Not at end of list, look to see if channel is "*" */
		else if (!strcmp(channel, "*"))
			strmcpy(channel, get_channel_by_refnum(0), 256);

		/* If there is still no channel, stop */
		if (!*channel)
			return (booya ? booya : m_strdup(empty_string));

		if ((key = channel_key(channel)))
		{
			if (booya)
			    malloc_strcat(&booya, " ");
			malloc_strcat(&booya, (key || *key) ? key : "*");
		}
	}
	while (strmcpy(channel, new_next_arg(word, &word), 256));

	if (!booya)
		RETURN_EMPTY;

	return (booya);
}

static
#ifdef __STDC__
char    *function_revw (char *words)
#else
char    *function_revw (words)
char    *words;
#endif
{
	char	*booya = (char *)new_malloc(strlen(words) + 1);
	*booya = 0;

	while (words && *words)
	{
		if (*booya)
			strcat(booya, " ");
		strcat(booya, last_arg(&words));
	}

	return booya;
}

static
#ifdef __STDC__
char    *function_reverse(char *words)
#else
char    *function_reverse(words)
char    *words;
#endif
{
	int     length = strlen(words);
	char    *booya = (char *) 0;
	int     x = 0;

	booya = (char *)new_malloc(length+1);
	for(length--; length >= 0; length--,x++)
		booya[x] = words[length];
	booya[x]='\0';
	return (booya);
}

static
#ifdef __STDC__
char    *function_jot(char *word)
#else
char    *function_jot(word)
char    *word;
#endif
{
	int     start = 0;
	int     stop = 0;
	int     interval = 1;
	int     counter;
	char    booya[4096];
	char    *foobar = (char *) 0;
	char    *foob = (char *) 0;

	bzero(booya, 4096);

	if ((foobar = new_next_arg(word, &word)))
	{
		start = atoi(foobar);
		if ((foobar = new_next_arg(word, &word)))
		{
			stop = atoi(foobar);
			if ((foobar = new_next_arg(word, &word)))
				interval = atoi(foobar);
			else
				interval = 1;
		}
		else
			return m_strdup(empty_string);
	}
	else
		return m_strdup(empty_string);

	if (interval < 0) 
		interval = -interval;
	if (start < stop)
		for (counter = start;counter <= stop;counter+=interval)
		{
			if (booya && *booya)
				sprintf(booya, "%s %d",booya,counter);
			else
				sprintf(booya, "%d", counter);
		}
	else
		for (counter = start;counter >= stop;counter-=interval)
		{
			if (booya && *booya)
				sprintf(booya, "%s %d",booya,counter);
			else
				sprintf(booya, "%d", counter);
		}
	booya[4095] = '\0';
	malloc_strcpy(&foob, booya);
	return (foob);
}

#ifdef __STDC__
extern char    *function_shift(char *word)
#else
extern char    *function_shift(word)
char	*word;
#endif
{
	char    *value = (char *) 0;
	char    *var    = (char *) 0;
	char	*booya 	= (char *) 0;
	char	*blah;
	int     old_display = window_display;
	char    *placeholder;

	GET_STR_ARG(var, word);

	if (word && *word)
	{
		malloc_strcpy(&value, var);
		malloc_strcat(&value, word);
		var = (char *) 0;
	}
	else
		value = find_inline(var);

	if (!value && !*value)
		RETURN_EMPTY;

	placeholder = value;
	blah = new_next_arg(value, &value);
	malloc_strcpy(&booya, blah);
	if (var)
	{
		window_display = 0;
		add_alias(VAR_ALIAS, var, value);
		window_display = old_display;
	}
	new_free(&placeholder);
	RETURN_STR(booya);
}

#ifdef __STDC__
extern char    *function_unshift(char *word)
#else
extern char    *function_unshift(word)
char	*word;
#endif
{
	char    *value = (char *) 0;
	char    *var    = (char *) 0;
	char	*booya  = (char *) 0;
	int     old_display = window_display;

	GET_STR_ARG(var, word);
	value = find_inline(var);
	if (!word || !*word)
		return value;

	malloc_strcat(&booya, word);
	if (value && *value)
		m_3cat(&booya, " ", value);

	window_display = 0;
	add_alias(VAR_ALIAS, var, booya);
	window_display = old_display;
	new_free(&value);
	RETURN_STR(booya);
}

#ifdef __STDC__
extern char    *function_push(char *word)
#else
extern char    *function_push(word)
char	*word;
#endif
{
	char    *value = (char *) 0;
	char    *var    = (char *) 0;
	int     old_display = window_display;

	GET_STR_ARG(var, word);
	value = find_inline(var);
	if (value && *value)
		malloc_strcat(&value, " ");
	malloc_strcat(&value, word);
	window_display = 0;
	add_alias(VAR_ALIAS, var, value);
	window_display = old_display;
	RETURN_STR(value);
}

#ifdef __STDC__
extern char	*function_pop(char *word)
#else
extern char	*function_pop(word)
char	*word;
#endif
{
	char *value	= (char *) 0;
	char *var	= (char *) 0;
	char *pointer	= (char *) 0;
	int   old_display = window_display;
	char *blech     = (char *) 0;

	GET_STR_ARG(var, word);

	if (word && *word)
	{
		pointer = rindex(word, ' ');
		return m_strdup(pointer ? pointer : word);
	}

	value = find_inline(var);
	if (!value || !*value)
	{
		new_free(&value);
		return m_strdup(empty_string);
	}

	pointer = rindex(value, ' ');
	if (!pointer)
	{
		window_display = 0;
		add_alias(VAR_ALIAS, var, empty_string); /* dont forget this! */
		window_display = old_display;
		return value;	/* one word -- return it */
	}

	*pointer++ = '\0';
	window_display = 0;
	add_alias(VAR_ALIAS, var, value);
	window_display = old_display;

	/* because pointer points to value, we *must* make a copy of it
	 * *before* we free value! (And we cant forget to free value, either) */
	blech = m_strdup(pointer);
	new_free(&value);
	RETURN_STR(blech);
}


/* Search and replace function --
   Usage:   $sar(c/search/replace/data)
   Commands:
		r - treat data as a variable name and 
		    return the replaced data to the variable
		g - Replace all instances, not just the first one
   The delimiter may be any character that is not a command (typically /)
   The delimiter MUST be the first character after the command
   Returns empty string on error
*/
/* This needs a good re-working to be more efficient. */
static
#ifdef __STDC__
char    *function_sar(char *word)
#else
char    *function_sar(word)
char    *word;
#endif
{
	char    delimiter;
	char	*pointer	= (char *) 0;
	char    *search         = (char *) 0;
	char    *replace        = (char *) 0;
	char    *data		= (char *) 0;
	char	*value		= (char *) 0;
	char	*booya		= (char *) 0;
	int	variable = 0,global = 0,searchlen,oldwindow = window_display;
	char 	*svalue;

	while (((*word == 'r') && (variable = 1)) || ((*word == 'g') && (global = 1)))
		word++;

	RETURN_IF_EMPTY(word);

	delimiter = *word;
	search = word + 1;
	if ((replace = index(search, delimiter)) == 0)
		return m_strdup(empty_string);

	*replace = 0;
	replace++;
	if ((data = index(replace,delimiter)) == 0)
		return m_strdup(empty_string);

	*data++ = '\0';

	value = (variable == 1) ? find_inline(data) : m_strdup(data);

	if ((!value) || (!*value))
	{
		if (value)
			new_free(&value);
		RETURN_EMPTY;
	}
	svalue = value;

	pointer = value;
	searchlen = strlen(search) - 1;

	if (global)
	{
		while ((pointer = strstr(pointer,search)) != (char *) 0)
		{
			pointer[0] = pointer[searchlen] = 0;
			pointer += searchlen + 1;
			malloc_strcat(&booya, value);
			malloc_strcat(&booya, replace);
			value = pointer;
		}
	} 
	else
	{
		if ((pointer = strstr(pointer,search)) != (char *) 0)
		{
			pointer[0] = pointer[searchlen] = 0;
			pointer += searchlen + 1;
			malloc_strcat(&booya, value);
			malloc_strcat(&booya, replace);
			value = pointer;
		}
	}

	malloc_strcat(&booya, value);
	if (variable == 1) 
	{
		window_display = 0;
		add_alias(VAR_ALIAS, data, booya);
		window_display = oldwindow;
	}
	new_free(&svalue);
	return (booya);
}

static
#ifdef __STDC__
char    *function_center(char *word)
#else
char    *function_center(word)
char    *word;
#endif
{
	int	length,pad,width;
	char 	*padc;

	if (!word || !*word)
		return m_strdup(empty_string);

	width = atoi(new_next_arg(word, &word));
	length = strlen(word);
	
	if ((pad = width - length) < 0)
		return (word);

	pad /= 2;
	padc = (char *)new_malloc(width+1);
	memset(padc, ' ', pad);
	padc[pad] = '\0';

	return strcat(padc, word);
}

static
#ifdef __STDC__
char    *function_split(char *word)
#else
char    *function_split(word)
char    *word;
#endif
{
	char	*chrs;
	char	*pointer;

	chrs = next_arg(word, &word);
	pointer = word;
	while ((pointer = sindex(pointer,chrs)))
		*pointer++ = ' ';

	return m_strdup(word);
}

static
#ifdef __STDC__
char *function_chr(char *word)
#else
char *function_chr(word)
char *word;
#endif
{
	char aboo[BIG_BUFFER_SIZE];
	char *ack = aboo;
	char *blah;

	while ((blah = next_arg(word, &word)))
		*ack++ = (char)atoi(blah);

	*ack = '\0';
	return m_strdup(aboo);
}

static
#ifdef __STDC__
char *function_ascii(char *word)
#else
char *function_ascii(word)
char *word;
#endif
{
	char aboo[BIG_BUFFER_SIZE];

	if (!word || !*word)
		return m_strdup (empty_string); /* no input, no output */

	sprintf(aboo, "%d", *word);
	while (*++word)
		sprintf(aboo, "%s %d", aboo, *word);

	return m_strdup(aboo);
}

static
#ifdef __STDC__
char	*function_which(char *word)
#else
char	*function_which(word)
char	*word;
#endif
{
	char *file1;
	FILE *fp;

	GET_STR_ARG(file1, word);
	file1 = m_strdup(file1);

	if ((fp = uzfopen (&file1, (word && *word) ? word : get_string_var(LOAD_PATH_VAR))))
	{
		fclose (fp);
		return (file1);
	}
	else
	{
		new_free(&file1);
		RETURN_EMPTY;
	}
}


static
#ifdef __STDC__
char *function_isalpha(char *words)
#else
char *function_isalpha(words)
char *words;
#endif
{
	if (((*words >= 'a') && (*words <= 'z')) ||
	    ((*words >= 'A') && (*words <= 'Z')))
		RETURN_INT(1)
	else
		RETURN_INT(0)
}

static
#ifdef __STDC__
char *function_isdigit(char *words)
#else
char *function_isdigit(words)
char *words;
#endif
{
	if (((*words >= '0') && (*words <= '9')) ||
	    ((*words == '-') && ((*(words+1) >= '0') || (*(words+1) <= '9'))))
		RETURN_INT(1)
	else
		RETURN_INT(0)
}

static
#ifdef __STDC__
char *function_open (char *words)
#else
char *function_open (words)
char *words;
#endif
{
	char *filename;
	GET_STR_ARG(filename, words);

	if (words && *words && toupper(*words) == 'R')
		RETURN_INT(open_file_for_read(filename))
	else if (words && *words && toupper(*words) == 'W')
		RETURN_INT(open_file_for_write(filename))
	else
		RETURN_EMPTY;
}

static
#ifdef __STDC__
char *function_close (char *words)
#else
char *function_close (words)
char *words;
#endif
{
	RETURN_IF_EMPTY(words);
	RETURN_INT(file_close(atoi(new_next_arg(words, &words))));
}	

static
#ifdef __STDC__
char *function_write (char *words)
#else
char *function_write (words)
char *words;
#endif
{
	char *fdc;
	GET_STR_ARG(fdc, words);
	RETURN_INT(file_write(atoi(fdc), words));
}

static
#ifdef __STDC__
char *function_read (char *words)
#else
char *function_read (words)
char *words;
#endif
{
	RETURN_IF_EMPTY(words);
	return file_read (atoi(new_next_arg(words, &words)));
}

static
#ifdef __STDC__
char *function_eof (char *words)
#else
char *function_eof (words)
char *words;
#endif
{
	RETURN_IF_EMPTY(words);
	RETURN_INT(file_eof(atoi(new_next_arg(words, &words))));
}


static
#ifdef __STDC__
char *function_iptoname (char *words)
#else
char *function_iptoname (words)
char *words;
#endif
{
	RETURN_STR(ip_to_host(words));
}

static
#ifdef __STDC__
char *function_nametoip (char *words)
#else
char *function_nametoip (words)
char *words;
#endif
{
	RETURN_STR(host_to_ip(words));
}

static
#ifdef __STDC__
char *function_convert (char *words)
#else
char *function_convert (words)
char *words;
#endif
{
	RETURN_STR(one_to_another(words));
}

static
#ifdef __STDC__
char *function_translate (char *words)
#else
char *function_translate (words)
char *words;
#endif
{
	char *	oldc, 
	     *	newc, 
	     *	text,
	     *	ptr,
		delim;
	int 	size_old, 
		size_new,
		x;

	RETURN_IF_EMPTY(words);

	oldc = words;
	/* First character can be a slash.  If it is, we just skip over it */
	delim = *oldc++;
	newc = strchr(oldc, delim);

	if (newc)
	{
		text = strchr(++newc, delim);

		if (newc == oldc)
			RETURN_EMPTY;

		if (!text)
			RETURN_EMPTY;
		*text++ = '\0';

		if (newc == text)
		{
			*newc = '\0';
			newc = empty_string;
		}
		else
			newc[-1] = 0;

		/* this is cheating, but oh well, >;-) */
		text = m_strdup(text);

		size_new = strlen(newc);
		size_old = strlen(oldc);

		for (ptr = text; ptr && *ptr; ptr++)
		{
			for (x = 0;x < size_old;(void)x++)
			{
				if (*ptr == oldc[x])
				{
					/* Check to make sure we arent
					   just eliminating the character.
					   If we arent, put in the new char,
					   otherwise strcpy it away */
					if (size_new)
						*ptr = newc[(x<size_new)?x:size_new-1];
					else
						strcpy (ptr, ptr+1);
					break;
				}
			}
		}
		return text;
	}
	else
		/* No text in, no text out */
		RETURN_EMPTY;
}

static
#ifdef __STDC__
char *function_server_version (char *word)
#else
char *function_server_version (word)
char *word
#endif
{
	int servnum;
	int version;

	servnum = ((word && *word) ? atoi(next_arg(word, &word)) : primary_server);

	if (servnum > number_of_servers)
		RETURN_STR("unknown");

	version = get_server_version(servnum);

#ifdef COMPAT_27
	if (version == Server2_7)		RETURN_STR("2.7");
	else
#endif
	if (version == Server2_8) 		RETURN_STR("2.8");
	else if (version == Server_u2_8) 	RETURN_STR("u2.8");
	else if (version == Server_u2_9) 	RETURN_STR("u2.9");
	else if (version == Server2_9) 		RETURN_STR("2.9");
	else if (version == Server_u3_0)	RETURN_STR("u3.0");

	RETURN_STR("Unknown");
}

static
#ifdef __STDC__
char *function_unlink(char *words)
#else
char *function_unlink (words)
char *words
#endif
{
	char *filename;
	int 	failure = 0;

	while ((filename = next_arg(words, &words)))
		failure -= unlink(filename);	

	RETURN_INT(failure);
}

static
#ifdef __STDC__
char *function_rename (char *words)
#else
char *function_rename (words)
char *words
#endif
{
	char *	filename1, 
	     *	filename2;
	int 	failure = 0;

	filename1 = next_arg(words, &words);
	filename2 = next_arg(words, &words);
	failure = filename2 ? rename(filename1, filename2) : 0;

	RETURN_INT(failure);
}

#ifdef __STDC__
static char *function_rmdir (char *words)
#else
static char *function_rmdir (words)
char *words
#endif
{
	int 	failure = 0;

	while (words && *words && (failure -= rmdir(next_arg(words, &words))))
		;

	RETURN_INT(failure);
}

#ifdef __STDC__
static char *function_truncate (char *words)
#else
static char *function_truncate (words)
char *words
#endif
{
	int		num = 0;
	double		value = 0;
	char		buffer[BIG_BUFFER_SIZE],
			format[1024];

	GET_INT_ARG(num, words);
	GET_FLOAT_ARG(value, words);

	if (num < 0)
	{
		float foo;
		int end;

		sprintf(format, "%%.%de", -num-1);
		sprintf(buffer, format, value);
		foo = atof(buffer);
		sprintf(buffer, "%f", foo);
		end = strlen(buffer) - 1;
		if (end == 0)
			RETURN_EMPTY;
		while (buffer[end] == '0')
			end--;
		if (buffer[end] == '.')
			end--;
		buffer[end+1] = 0;
	}
	else if (num > 0)
	{
		sprintf(format, "%%10.%dlf", num);
		sprintf(buffer, format, value);
	}
	else
		RETURN_EMPTY;

	while (*buffer == ' ')
		strcpy(buffer, buffer+1);

	RETURN_STR(buffer);
}
